"""Simple library functions which don't depend on adverbs"""
import __builtin__

import syntax
import syntax_helpers

import numpy as np 
import prims 
from decorators import macro, staged_macro, jit

import array_type
from array_type import ArrayT
import core_types 
from core_types import Int8, Int16, Int32, Int64
from core_types import Float32, Float64
from core_types import UInt8, UInt16, UInt32, UInt64
from core_types import Bool
from syntax import Map, AllPairs, Reduce, Scan, IndexMap, IndexReduce
from syntax import DelayUntilTyped, Cast, Range, Attribute, AllocArray
from syntax import Tuple, TupleProj, ArrayView, Index
from syntax_helpers import zero_i64, one_i64, one_i32, const_int 
from tuple_type import empty_tuple_t, TupleT 

@jit 
def identity(x):
  return x

@macro 
def parfor(shape, fn):
  return syntax.ParFor(fn = fn, shape = shape)

@staged_macro("axis")
def map(f, *args, **kwds):
  axis = kwds.get('axis', syntax_helpers.zero_i64)
  return Map(fn = f, args = args, axis = axis)

each = map

@staged_macro("axis") 
def allpairs(f, x, y, axis = 0):
  return AllPairs(fn = f, args = (x,y), axis = axis)

@staged_macro("axis")
def reduce(f, *args, **kwds):
  axis = kwds.get('axis', syntax_helpers.none)
  init = kwds.get('init', syntax_helpers.none)
  import ast_conversion
  ident = ast_conversion.translate_function_value(identity)
  return Reduce(fn = ident, 
                combine = f, 
                args = args,
                init = init,
                axis = axis)

@staged_macro("axis")
def scan(f, *args, **kwds):
  axis = kwds.get('axis', syntax_helpers.zero_i64)
  init = kwds.get('init', syntax_helpers.none)
  import ast_conversion
  ident = ast_conversion.translate_function_value(identity)
  return Scan(fn = ident,  
                     combine = f,
                     emit = ident, 
                     args = args,
                     init = init,
                     axis = axis)

@macro
def imap(fn, shape):
  return IndexMap(shape = shape, fn = fn)


@macro
def ireduce(combine, shape, map_fn = identity, init = None):
  return IndexReduce(fn = map_fn, combine=combine, shape = shape, init = init)

@jit 
def _tuple_from_args(*args):
  return args

@macro
def zip(*args):
  import ast_conversion 
  elt_tupler = ast_conversion.translate_function_value(_tuple_from_args)
  return Map(fn = elt_tupler, args = args)

@macro 
def int8(x):
  return Cast(x, type = Int8) 

@macro 
def int16(x):
  return Cast(x, type = Int16) 

@macro 
def int32(x):
  return Cast(x, type = Int32) 

@macro 
def int64(x):
  return Cast(x, type = Int64) 


@macro 
def uint8(x):
  return Cast(x, type = UInt8) 

@macro 
def uint16(x):
  return Cast(x, type = UInt16) 

@macro 
def uint32(x):
  return Cast(x, type = UInt32) 

@macro 
def uint64(x):
  return Cast(x, type = UInt64)

uint = uint64 

@macro 
def float32(x):
  return Cast(x, type = Float32)

@macro 
def float64(x):
  return Cast(x, type = Float64)

@macro 
def bool(x):
  return Cast(x, type = Bool)

@jit
def real(x):
  """
  For now we don't have complex types, so real is just the identity function
  """
  return x 

@macro 
def _builtin_tuple(x):
  def typed_tuple(xt):
    if isinstance(xt.type, TupleT):
      return xt 
    else:
      assert isinstance(xt.type, ArrayT), "Can't create type from %s" % (xt.type,)
      assert isinstance(xt, syntax.Array), "Can only create tuple from array of const length"
      elt_types = [e.type for e in xt.elts]
      import tuple_type
      tuple_t = tuple_type.make_tuple_type(elt_types)
      return syntax.Tuple(xt.elts, type = tuple_t)
  return DelayUntilTyped(x, typed_tuple)

@macro 
def alen(x):
  def typed_alen(xt):
    if isinstance(xt.type, ArrayT):
      shape = Attribute(xt, 'shape', type = xt.type.shape_t)
      return TupleProj(shape, 0, type = Int64)
    else:
      assert isinstance(xt.type, TupleT), "Can't get 'len' of object of type %s" % xt.type 
      return const_int(len(xt.type.elt_types))
  return DelayUntilTyped(x, typed_alen)

@macro 
def shape(x):
  def typed_shape(xt):
    if isinstance(xt.type, ArrayT):
      return Attribute(xt, 'shape', type = xt.type.shape_t)
    else:
      return Tuple((), type = empty_tuple_t)
    
  return DelayUntilTyped(x, typed_shape)

@jit 
def sum(x, axis = None):
  return reduce(prims.add, x, init = 0, axis = axis)

@jit 
def prod(x, axis=None):
  return reduce(prims.multiply, x, init=1, axis = axis)

@jit 
def mean(x, axis = None):
  return sum(x, axis = axis) / x.shape[0]

@jit 
def cumsum(x, axis = None):
  return scan(prims.add, x, axis = axis)

@jit 
def cumprod(x, axis = None):
  return scan(prims.multiply, x, axis = axis)

@jit 
def or_(x, y):
  return x or y

@jit 
def and_(x, y):
  return x and y

@jit 
def any(x, axis=None):
  return reduce(or_, x, axis = axis, init = False)

@jit
def all(x, axis = None):
  return reduce(and_, x, axis = axis, init = True)


@jit 
def diff(x):
  """
  TODO:
    - axis selection
    - preserve size by filling with zeros
    - allow n'th differences by recursion
  """
  return x[1:] - x[:-1]

@jit 
def dot(x,y):
  return sum(x*y)

@jit 
def min(x, axis = None):
  return reduce(prims.minimum, x, axis = axis)

@jit 
def _builtin_min(x, y = None):
  if y is None:
    return min(x)
  else:
    return prims.minimum(x,y)

@jit 
def max(x, axis = None):
  return reduce(prims.maximum, x, axis = axis)

@jit
def _builtin_max(x, y = None):
  if y is None:
    return max(x)
  else:
    return prims.maximum(x,y)
  
@macro
def arange(n, *xs):
  count = __builtin__.len(xs)
  assert 0 <= count <= 2, "Too many args for range: %s" % ((n,) + tuple(xs))
  if count == 0:
    return Range(syntax_helpers.zero_i64, n, syntax_helpers.one_i64)
  elif count == 1:
    return Range(n, xs[0], syntax_helpers.one_i64)  
  else:
    return Range(n, xs[0], xs[1])
 
 
@macro
def empty(shape, dtype = float64):

  def typed_empty(shape, dtype):
    # HACK! 
    # In addition to TypeValue, allow casting functions 
    # to be treated as dtypes 
    if isinstance(dtype, syntax.Fn):
      assert len(dtype.body) == 1
      stmt = dtype.body[0]
      assert stmt.__class__ is syntax.Return 
      expr = stmt.value 
      assert expr.__class__ is syntax.Cast
      elt_t = expr.type 
    elif isinstance(dtype, syntax.TypedFn):
      elt_t = dtype.return_type
    else:
      assert isinstance(dtype.type, core_types.TypeValueT), \
         "Invalid dtype %s " % (dtype,)
      elt_t = dtype.type.type 
    assert isinstance(elt_t, core_types.ScalarT), \
       "Array element type %s must be scalar" % (elt_t,)  
    if isinstance(shape, core_types.ScalarT):
      shape = Tuple((shape,))
    rank = len(shape.type.elt_types)
    arr_t = array_type.make_array_type(elt_t, rank)
    return AllocArray(shape = shape, elt_type = elt_t, type = arr_t)
  return DelayUntilTyped(values=(shape,dtype), fn = typed_empty) 

@jit 
def empty_like(x, dtype = None):
  if dtype is None:
    return empty(x.shape, x.dtype)
  else:
    return empty(x.shape, dtype)
  
@jit 
def zeros(shape, dtype = float64):
  zero = dtype(0)
  return imap(lambda _: zero, shape)


@jit
def zeros_like(x, dtype = None):
  if dtype is None:
    dtype = x.dtype
  return zeros(x.shape, dtype)

@jit
def ones(shape, dtype = float64):
  one = dtype(1)
  return imap(lambda _: one, shape)

@jit
def ones_like(x, dtype = None):
  if dtype is None:
    dtype = x.dtype
  return ones(x.shape)

@macro
def transpose(x):
  def typed_transpose(xt):
    if isinstance(xt.type, ArrayT) and xt.type.rank > 1:
      shape = Attribute(xt, 'shape', type = xt.type.shape_t)
      strides = Attribute(xt, 'strides', type = xt.type.strides_t)
      data = Attribute(xt, 'data', type = xt.type.ptr_t)
      size = Attribute(xt, 'size', type = Int64)
      offset = Attribute(xt, 'offset', type = Int64)
      ndims = xt.type.rank 
      shape_elts = [TupleProj(shape, i, type = Int64) 
                               for i in xrange(ndims)]
      stride_elts = [TupleProj(strides, i, type = Int64) 
                                 for i in xrange(ndims)]
      new_shape = Tuple(tuple(reversed(shape_elts)))
      new_strides = Tuple(tuple(reversed(stride_elts)))
      return ArrayView(data, 
                       new_shape, 
                       new_strides, 
                       offset, 
                       size, 
                       type = xt.type)
    else:
      return xt 
  return DelayUntilTyped(x, typed_transpose)   

@macro 
def ravel(x):
  return syntax.Ravel(x)
  
@macro 
def reshape(x):
  return syntax.Reshape(x)

@macro 
def elt_type(x):
  def typed_elt_type(xt):
    return syntax.TypeValue(array_type.elt_type(xt.type))
  return DelayUntilTyped(x, typed_elt_type)

@macro
def itemsize(x):
  def typed_itemsize(xt):
    return const_int(array_type.elt_type(xt.type).nbytes)
  return DelayUntilTyped(x, typed_itemsize)

@macro 
def rank(x):
  def typed_rank(xt):
    return const_int(xt.type.rank) 
  return DelayUntilTyped(x, typed_rank)
@macro 
def size(x):
  def typed_size(xt):
    if isinstance(xt.type, array_type.ArrayT):
      return Attribute(xt, 'size', type = Int64)
    else:
      return const_int(1)
  return DelayUntilTyped(x, typed_size)

@jit 
def fill(x, v):
  for i in range(len(x)):
    x[i] = v 
    
    
@jit
def argmax(x):
  """
  Currently assumes axis=None
  TODO: 
    - Support axis arguments
    - use IndexReduce instead of explicit loop
  
      def argmax_map(curr_idx):
        return curr_idx, x[curr_idx]
  
      def argmax_combine((i1,v1), (i2,v2)):
        if v1 > v2:
          return (i1,v1)
        else:
          return (i2,v2)
    
      return ireduce(combine=argmin_combine, shape=x.shape, map_fn=argmin_map, init = (0,x[0]))
  """
  bestval = x[0]
  bestidx = 0
  for i in xrange(1, len(x)):
    currval = x[i]
    if currval > bestval:
      bestval = currval
      bestidx = i
  return bestidx 

@jit
def argmin(x):
  """
  Currently assumes axis=None
  TODO: 
    - Support axis arguments
    - use IndexReduce instead of explicit loop
  
      def argmin_map(curr_idx):
        return curr_idx, x[curr_idx]
  
      def argmin_combine((i1,v1), (i2,v2)):
        if v1 < v2:
          return (i1,v1)
        else:
          return (i2,v2)
    
      return ireduce(combine=argmin_combine, shape=x.shape, map_fn=argmin_map, init = (0,x[0]))
  """
  bestval = x[0]
  bestidx = 0
  for i in xrange(1, len(x)):
    currval = x[i]
    if currval < bestval:
      bestval = currval
      bestidx = i
  return bestidx 

@jit
def copy(x):
  return [xi for xi in x]


@jit
def pmap1(f, x, w = 3):
  n = x.shape[0]
  def local_apply(i):
    lower = __builtins__.max(i-w/2, 0)
    upper = __builtins__.min(i+w/2+1, n)
    elts = x[lower:upper]
    return f(elts)
  return imap(local_apply, n)

@jit  
def pmap2(f, x, width = (3,3)):
  """
  Patch-map where the function can accept both interior windows
  and smaller border windows
  """
  width_x, width_y = width 
  n_rows, n_cols = x.shape
  hx = width_x / 2
  hy = width_y / 2
  def local_apply((i,j)):
    lx = __builtins__.max(i-hx, 0)
    ux = __builtins__.min(i+hx+1, n_rows)
    ly = __builtins__.max(j-hy, 0)
    uy = __builtins__.min(j+hy+1, n_cols)
    return f(x[lx:ux, ly:uy])
    
  return imap(local_apply, x.shape)
    
@jit  
def pmap2_trim(f, x, width = (3,3), step = (1,1)):
  """
  Patch-map over interior windows, ignoring the border
  """
  width_x, width_y = width 
  step_x, step_y = step 
  n_rows, n_cols = x.shape
  hx = width_x / 2
  hy = width_y / 2
  return [[f(x[i-hx:i+hx+1, j-hy:j+hy+1]) 
           for j in np.arange(hx, n_cols-hx, step_x)] 
           for i in np.arange(hy, n_rows-hy, step_y)]






import args
from   args import ActualArgs
import core_types
from  node import Node
from syntax_stmts import * 
from syntax_expr import *
from syntax_array_expr import * 
from syntax_adverbs import * 


class DelayUntilTyped(Expr):
  """
  Once the list of values has been annotated with locally inferred types, 
  pass them to the given function to construct a final expression 
  """
  _members = ['values', 'fn']
 
  def node_init(self):
    if isinstance(self.values, list):
      self.values = tuple(self.values)
    elif not isinstance(self.values, tuple):
      self.values = (self.values,)
    
  def children(self):
    return self.values

class TypeValue(Expr):
  """
  Value materialization of a type 
  """
  
  _members = ['type_value']
  
  def node_init(self):
    if self.type is None:
      self.type = core_types.TypeValueT(self.type_value)
    assert isinstance(self.type, core_types.TypeValueT)
    assert self.type.type 
    
class Fn(Expr):
  """
  Function definition.
  A top-level function can have references to python values from its enclosing
  scope, which are stored in the 'python_refs' field.

  A nested function, on the other hand, might refer to some variables from its
  enclosing Parakeet scope, whose original names are stored in
  'parakeet_nonlocals'
  """

  _members = ['name', 'args', 'body', 'python_refs', 'parakeet_nonlocals']
  registry = {}

  def __str__(self):
    return "Fn(%s)" % self.name

  def __repr__(self):
    return "def %s(%s):%s" % (self.name, self.args, block_to_str(self.body))

  def __hash__(self):
    return hash(self.name)

  def node_init(self):
    assert isinstance(self.name, str), \
        "Expected string for fn name, got %s" % self.name
    assert isinstance(self.args, args.FormalArgs), \
        "Expected arguments to fn to be FormalArgs object, got %s" % self.args
    assert isinstance(self.body, list), \
        "Expected body of fn to be list of statements, got " + str(self.body)

    import closure_type
    self.type = closure_type.make_closure_type(self, ())
    self.registry[self.name] = self

  def python_nonlocals(self):
    if self.python_refs:
      return [ref.deref() for ref in self.python_refs]
    else:
      return []

  def children(self):
    return ()

################################################################################
#
#  Constructs below here are only used in the typed representation
#
################################################################################


class TypedFn(Expr):
  """
  The body of a TypedFn should contain Expr nodes which have been extended with
  a 'type' attribute
  """

  _members = ['name',
              'arg_names',
              'body',
              'input_types',
              'return_type',
              'type_env',
              # these last two get filled by
              # transformation/optimizations later
              'copied_by',
              'version',]

  registry = {}
  max_version = {}
  def next_version(self, name):
    n = self.max_version[name] + 1
    self.max_version[name] = n
    return n

  def node_init(self):
    assert isinstance(self.body, list), \
        "Invalid body for typed function: %s" % (self.body,)
    assert isinstance(self.arg_names, (list, tuple)), \
        "Invalid typed function arguments: %s" % (self.arg_names,)
    assert isinstance(self.name, str), \
        "Invalid typed function name: %s" % (self.name,)

    if isinstance(self.input_types, list):
      self.input_types = tuple(self.input_types)

    assert isinstance(self.input_types, tuple), \
        "Invalid input types: %s" % (self.input_types,)
    assert isinstance(self.return_type, core_types.Type), \
        "Invalid return type: %s" % (self.return_type,)
    assert isinstance(self.type_env, dict), \
        "Invalid type environment: %s" % (self.type_env,)

    self.type = core_types.make_fn_type(self.input_types, self.return_type)

    if self.version is None:
      self.version = 0
      self.max_version[self.name] = 0

    registry_key = (self.name, self.version)
    assert registry_key not in self.registry, \
        "Typed function %s version %s already registered" % \
        (self.name, self.version)
    self.registry[registry_key] = self

  def __repr__(self):
    arg_strings = []

    for name in self.arg_names:
      arg_strings.append("%s : %s" % (name, self.type_env.get(name)))

    return "function %s(%s) => %s:%s" % \
           (self.name, ", ".join(arg_strings),
            self.return_type,
            block_to_str(self.body))

  def __str__(self):
    return "TypedFn(%s : %s => %s)" % (self.name, self.input_types, self.return_type)

  def __hash__(self):
    return hash(self.name)

  def children(self):
    return ()



import sys
import time 

import numpy as np
from nose.tools import nottest

import parakeet
from parakeet import ast_conversion, interp, type_conv, type_inference
from parakeet.run_function import specialize_and_compile

def run_local_functions(prefix, locals_dict = None):
  if locals_dict is None:
    last_frame = sys._getframe()
    locals_dict = last_frame.f_back.f_locals

  good = set([])
  # bad = set([])
  for k, test in locals_dict.iteritems():
    if k.startswith(prefix):
      print "Running %s..." % k
      try:
        test()
        print "\n --- %s passed\n" % k
        good.add(k)
      except:
        raise

  print "\n%d tests passed: %s\n" % (len(good), ", ".join(good))

@nottest
def run_local_tests(locals_dict = None):
  if locals_dict is None:
    last_frame = sys._getframe()
    locals_dict = last_frame.f_back.f_locals
  return run_local_functions("test_", locals_dict)

def eq(x,y):
  if isinstance(x, np.ndarray) and not isinstance(y, np.ndarray):
    return False
  if isinstance(y, np.ndarray):
    if isinstance(x, np.ndarray) and x.shape == y.shape:
      err = abs(np.mean(np.ravel(x) - np.ravel(y)))
      m = abs(np.mean(np.ravel(x)))
      if not np.all(np.ravel(x) == np.ravel(y)) and err/m > 0.000001:
        print "err:", err
        print "err/m:", err/m
        return False
      else:
        return True
  elif np.isscalar(x) and np.isnan(x):
    return np.isscalar(x) and np.isnan(y)
  else:
    return np.allclose(x,y)

def copy(x):
  if isinstance(x, np.ndarray):
    return x.copy()
  else:
    return x

def expect(fn, args, expected, valid_types = None):
  """
  Helper function used for testing, assert that Parakeet evaluates given code to
  the correct result
  """

  untyped, typed, compiled, all_args = specialize_and_compile(fn, args)

  # Abandoned interpreting untyped code due to addition of DelayUntilTyped
  # syntax nodes 
  #untyped_result = interp.eval_fn(untyped, all_args.transform(copy))
  #assert eq(untyped_result, expected), \
  #    "Expected %s but untyped fn returned  %s" % (expected, untyped_result)
  
  linear_args = untyped.args.linearize_without_defaults(all_args)


  typed_result = interp.eval_fn(typed, map(copy, linear_args))
  assert eq(typed_result, expected), \
      "Expected %s but typed fn returned %s" % (expected, typed_result)

  llvm_result = compiled(*linear_args)
  if valid_types is not None:
    if not isinstance(valid_types, (tuple, list)):
      valid_types = [valid_types]
    assert type(llvm_result) in valid_types, \
      "Expected result to have type in %s but got %s" % (valid_types, type(llvm_result))
      
  assert eq(llvm_result, expected), \
      "Expected %s but compiled fn returned %s" % (expected, llvm_result)
  

def expect_each(parakeet_fn, python_fn, inputs):
  for x in inputs:
    expect(parakeet_fn, [x], python_fn(x))

def expect_allpairs(parakeet_fn, python_fn, inputs):
  for x in inputs:
    for y in inputs:
      expect(parakeet_fn, [x,y], python_fn(x,y))

def return_type(fn, input_types):
  untyped_fundef = ast_conversion.translate_function_value(fn)
  closure_args = untyped_fundef.python_nonlocals()
  closure_arg_types = map(type_conv.typeof, closure_args)
  return type_inference.infer_return_type(untyped_fundef,
                                          closure_arg_types + input_types)

def expect_type(fn, input_types, output_type):
  actual = return_type(fn, input_types)
  assert actual == output_type, "Expected type %s, actual %s" % \
                                (output_type, actual)
@nottest
def timed_test(parakeet_fn, parakeet_args, python_fn, 
               python_args = None, min_speedup = None):
  if python_args is None:
    python_args = parakeet_args

  start = time.time()
  _ = parakeet_fn(*parakeet_args)
  parakeet_time_with_comp = time.time() - start 

  start = time.time()
  py_result = python_fn(*python_args)
  py_time = time.time() - start 

  start = time.time()
  parakeet_result = parakeet_fn(*parakeet_args)
  parakeet_time_no_comp = time.time() - start 

  print "Parakeet time (with compilation):", parakeet_time_with_comp
  print "Parakeet time (without compilation):", parakeet_time_no_comp
  print "Python time:", py_time 

  assert eq(parakeet_result, py_result), \
    "Expected %s but got %s" % (py_result, parakeet_result)
  if min_speedup is not None:
    assert py_time / parakeet_time_no_comp > min_speedup, \
        "Parakeet too slow: %.2f slowdown" % (parakeet_time_no_comp/py_time)



import parakeet 

from parakeet.core_types import Int64 
from parakeet.escape_analysis import EscapeAnalysis
from parakeet.syntax import Tuple, Var, TypedFn, Assign, Return, TupleProj 
from parakeet.syntax_helpers import zero_i64, one_i64   
from testing_helpers import run_local_tests
from parakeet.tuple_type import make_tuple_type
 

tuple_t = make_tuple_type((Int64, Int64))
nested_tuple_t = make_tuple_type((tuple_t, tuple_t))

a_var = Var("a", type = Int64)
b_var = Var("b", type = tuple_t)
c_var = Var("c", type = tuple_t)
d_var = Var("d", type= nested_tuple_t)
e_var = Var("e", type = nested_tuple_t)
body = [
  Assign(a_var, one_i64),  
  Assign(b_var, Tuple(elts = (zero_i64, a_var), type = tuple_t)), 
  Assign(c_var, Tuple(elts = (TupleProj(b_var, 0, type = Int64), a_var), 
                      type = tuple_t)), 
  Assign(d_var, Tuple(elts = (b_var, b_var), type = nested_tuple_t)), 
  Assign(e_var, Tuple(elts = (c_var, c_var), type = nested_tuple_t)), 
  Return(d_var) 
]
tenv = { "a": Int64, 
         "b": tuple_t, 
         "c": tuple_t, 
         "d": nested_tuple_t, 
         "e": nested_tuple_t
        }

fn = TypedFn(name = "test_escape_analysis", 
             type_env = tenv, 
             input_types = (), 
             arg_names = (), 
             body = body, 
             return_type = nested_tuple_t
            )

def test_escape_analysis():
  escape_analysis = EscapeAnalysis()
  escape_analysis.visit_fn(fn)
  assert "a" not in escape_analysis.may_escape, "Scalars can't escape function"
  assert "b" in escape_analysis.may_escape, "Nested tuples also escape!"
  assert "e" not in escape_analysis.may_escape
  
if __name__ == '__main__':
  run_local_tests()



from testing_helpers import expect, run_local_tests

def implicit_to_float(x):
  return x + 0.5

def test_implicit_to_float():
  expect(implicit_to_float, [1], 1.5)
  expect(implicit_to_float, [True], 1.5)

def implicit_to_bool(b):
  if b:
    return 10
  else:
    return -10

def test_implicit_to_bool():
  expect(implicit_to_bool, [1], 10)
  expect(implicit_to_bool, [2], 10)
  expect(implicit_to_bool, [0], -10)
  expect(implicit_to_bool, [1.0], 10)
  expect(implicit_to_bool, [2.0], 10)
  expect(implicit_to_bool, [0.0], -10)

if __name__ == '__main__':
    run_local_tests()



import prims
import subst
import syntax
import syntax_helpers
import transform

from array_type import ArrayT, SliceT 
from closure_type import ClosureT
from collect_vars import collect_var_names
from core_types import NoneT, ScalarT
from mutability_analysis import TypeBasedMutabilityAnalysis
from scoped_dict import ScopedDictionary
from syntax import AllocArray, Assign, ExprStmt, Adverb
from syntax import Const, Var, Tuple, TupleProj, Closure, ClosureElt, Cast
from syntax import Slice, Index, Array, ArrayView, Attribute, Struct
from syntax import PrimCall, Call, TypedFn, Fn
from syntax import AllPairs, Map, Reduce, Scan, IndexMap, IndexReduce
from syntax_helpers import collect_constants, is_one, is_zero, all_constants
from syntax_helpers import get_types, slice_none_t, const_int
from transform import Transform
from tuple_type import TupleT
from use_analysis import use_count


# classes of expressions known to have no side effects
# and to be unaffected by changes in mutable state as long
# as all their arguments are SSA variables or constants
#
# examples of unsafe expressions:
#  - Index: the underlying array is mutable, thus the expression depends on
#    any data modifications
#  - Call: Unless the function is known to contain only safe expressions it
#    might depend on mutable state or modify it itself

class Simplify(Transform):
  def __init__(self):
    transform.Transform.__init__(self)
    # associate var names with any immutable values
    # they are bound to
    self.bindings = {}

    # which expressions have already been computed
    # and stored in some variable?
    self.available_expressions = ScopedDictionary()

  def pre_apply(self, fn):
    ma = TypeBasedMutabilityAnalysis()

    # which types have elements that might
    # change between two accesses?
    self.mutable_types = ma.visit_fn(fn)
    self.use_counts = use_count(fn)

  def immutable_type(self, t):
    return t not in self.mutable_types

  def children(self, expr, allow_mutable = False):
    c = expr.__class__

    if c is Const or c is Var:
      return ()
    elif c is PrimCall or c is Closure:
      return expr.args
    elif c is ClosureElt:
      return (expr.closure,)
    elif c is Tuple:
      return expr.elts
    elif c is TupleProj:
      return (expr.tuple,)
    # WARNING: this is only valid
    # if attributes are immutable
    elif c is Attribute:
      return (expr.value,)
    elif c is Slice:
      return (expr.start, expr.stop, expr.step)
    elif c is Cast:
      return (expr.value,)
    elif c is Map or c is AllPairs or c is IndexMap:
      return expr.args
    elif c is Scan or c is Reduce or c is IndexReduce:
      args = tuple(expr.args)
      init = (expr.init,) if expr.init else ()
      return init + args
    elif c is Call:
      # assume all Calls might modify their arguments
      if allow_mutable or all(self.immutable(arg) for arg in expr.args):
        return expr.args
      else:
        return None

    if allow_mutable or self.immutable_type(expr.type):
      if c is Array:
        return expr.elts
      elif c is ArrayView:
        return (expr.data, expr.shape, expr.strides, expr.offset, expr.size)
      elif c is Struct:
        return expr.args
      elif c is AllocArray:
        return (expr.shape,)
      elif c is Attribute:
        return (expr.value,)
    return None

  def immutable(self, expr):
    return (isinstance(expr, (Const, Tuple, Adverb, Cast, Var, PrimCall)) and 
            (all(self.immutable(c) for c in expr.children())) or \
           (isinstance(expr, (Attribute, TupleProj)) and \
            isinstance(expr.type, (ScalarT, TupleT, SliceT))))
    
                      
  def temp(self, expr, name = None, use_count = 1):
    """
    Wrapper around Codegen.assign_temp which also updates bindings and
    use_counts
    """
    if self.is_simple(expr):
      return expr 
    else:
      new_var = self.assign_temp(expr, name = name)
      self.bindings[new_var.name] = expr
      self.use_counts[new_var.name] = use_count
      return new_var

  def transform_expr(self, expr):
    if not self.is_simple(expr):
      stored = self.available_expressions.get(expr)
      if stored is not None:
        return stored
    return Transform.transform_expr(self, expr)

  def transform_Var(self, expr):
    name = expr.name
    prev_expr = expr
    while name in self.bindings:
      prev_expr = expr 
      expr = self.bindings[name]
      if expr.__class__ is Var:
        name = expr.name
      else:
        break
    c = expr.__class__ 
    if c is Var or c is Const:
      return expr
    else:
      return prev_expr
  
  def transform_Cast(self, expr):
    
    v = self.transform_expr(expr.value)
    if v.type == expr.type:
      return v
    elif v.__class__ is Const and isinstance(expr.type, ScalarT):
      return Const(expr.type.dtype.type(v.value), type = expr.type)
    elif self.is_simple(v):
      expr.value = v
      return expr
    else:
      expr.value = self.assign_temp(v)
      return expr

  def transform_Attribute(self, expr):
    v = self.transform_expr(expr.value)
    if v.__class__ is Var and v.name in self.bindings:
      stored_v = self.bindings[v.name]
      c = stored_v.__class__
      if c is Var or c is Struct:
        v = stored_v
      elif c is ArrayView:
        if expr.name == 'shape':
          return self.transform_expr(stored_v.shape)
        elif expr.name == 'strides':
          return self.transform_expr(stored_v.strides)
        elif expr.name == 'data':
          return self.transform_expr(stored_v.data)
      elif c is AllocArray:
        if expr.name == 'shape':
          return self.transform_expr(stored_v.shape)
    if v.__class__ is Struct:
      idx = v.type.field_pos(expr.name)
      return v.args[idx]
    elif v.__class__ is not Var:
      v = self.temp(v, "struct")
    if expr.value == v:
      return expr
    else:
      return Attribute(value = v, name = expr.name, type = expr.type)
    
  
  def transform_Closure(self, expr):
    expr.args = tuple(self.transform_args(expr.args))
    return expr

  def transform_Tuple(self, expr):
    expr.elts = tuple( self.transform_args(expr.elts))
    return expr

  def transform_TupleProj(self, expr):
    idx = expr.index
    assert isinstance(idx, int), \
        "TupleProj index must be an integer, got: " + str(idx)
    new_tuple = self.transform_expr(expr.tuple)

    if new_tuple.__class__ is Var and new_tuple.name in self.bindings:
      tuple_expr = self.bindings[new_tuple.name]
      if tuple_expr.__class__ is Tuple:
        return tuple_expr.elts[idx]
      elif tuple_expr.__class__ is Struct:
        return tuple_expr.args[idx]

    if not self.is_simple(new_tuple):
      new_tuple = self.assign_temp(new_tuple, "tuple")
    expr.tuple = new_tuple
    return expr

  def transform_ClosureElt(self, expr):
    idx = expr.index
    assert isinstance(idx, int), \
        "ClosureElt index must be an integer, got: " + str(idx)
    new_closure = self.transform_expr(expr.closure)
    if new_closure.__class__ is Var and new_closure.name in self.bindings:
      closure_expr = self.bindings[new_closure.name]
      if closure_expr.__class__ is Closure:
        return closure_expr.args[idx]

    if not self.is_simple(new_closure):
      new_closure = self.assign_temp(new_closure, "closure")
    expr.closure = new_closure
    return expr

  def transform_Call(self, expr):
    fn = self.transform_expr(expr.fn)
    args = self.transform_args(expr.args)
    if fn.type.__class__ is ClosureT:
      closure_elts = self.closure_elts(fn)
      combined_args = closure_elts + args
      if fn.type.fn.__class__ is TypedFn:
        fn = fn.type.fn
      else:
        assert isinstance(fn.type.fn, Fn)
        import type_inference
        fn = type_inference.specialize(fn, get_types(combined_args))
      assert fn.return_type == expr.type
      return Call(fn, combined_args, type = fn.return_type)
    else:
      expr.fn = fn
      expr.args = args
      return expr

  def transform_arg(self, x, name = None):

    return self.temp(self.transform_expr(x), name = name)
  
  def transform_args(self, args):
    return [self.transform_arg(x) for x in args]

  def transform_Array(self, expr):
    expr.elts = tuple(self.transform_args(expr.elts))
    return expr

  def transform_Index(self, expr):
    expr.value = self.transform_expr(expr.value)
    expr.index = self.transform_expr(expr.index)
    if expr.value.__class__ is Array and expr.index.__class__ is Const:
      assert isinstance(expr.index.value, (int, long)) and \
             len(expr.value.elts) > expr.index.value

      return expr.value.elts[expr.index.value]
    if expr.value.__class__ is not Var:
      expr.value = self.temp(expr.value, "array")
    return expr

  def transform_Struct(self, expr):
    new_args = self.transform_args(expr.args)
    return syntax.Struct(new_args, type = expr.type)

  def transform_PrimCall(self, expr):
    args = self.transform_args(expr.args)
    prim = expr.prim
    if all_constants(args):
      return syntax.Const(value = prim.fn(*collect_constants(args)),
                          type = expr.type)
    elif prim == prims.add:
      x,y = args
      if is_zero(x):
        return y
      elif is_zero(y):
        return x
    elif prim == prims.multiply:
      x,y = args
      if is_one(x):
        return y
      elif is_one(y):
        return x
      elif is_zero(x):
        return x
      elif is_zero(y):
        return y
      
    elif prim == prims.divide:
      x,y = args 
      if is_one(y):
        return x
      
    elif prim == prims.power:
      x,y = args
      if is_one(y):
        return self.cast(x, expr.type)
      elif is_zero(y):
        return syntax_helpers.one(expr.type)
      elif y.__class__ is Const and y.value == 2:
        return self.cast(self.mul(x, x, "sqr"), expr.type)

    expr.args = args
    return expr 
  
  def transform_Reduce(self, expr):
    
    init = self.transform_expr(expr.init)
    if not self.is_simple(init):
      expr.init = self.assign_temp(init, 'init')
    else:
      expr.init = init
    expr.args = self.transform_args(expr.args)
    expr.fn = self.transform_expr(expr.fn)
    expr.combine = self.transform_expr(expr.combine)
    return expr  
  
  def temp_in_block(self, expr, block, name = None):
    """
    If we need a temporary variable not in the current top scope but in a
    particular block, then use this function. (this function also modifies the
    bindings dictionary)
    """
    if name is None:
      name = "temp"
    var = self.fresh_var(expr.type, name)
    block.append(Assign(var, expr))
    self.bindings[var.name] = expr
    return var

  def set_binding(self, name, value):
    assert value.__class__ is not Var or \
        value.name != name, \
        "Can't set name %s bound to itself" % name
    self.bindings[name] = value

  def bind_var(self, name, rhs):
    if rhs.__class__ is Var:
      old_val = self.bindings.get(rhs.name)
      if old_val and self.is_simple(old_val):
        self.set_binding(name, old_val)
      else:
        self.set_binding(name, rhs)
    else:
      self.set_binding(name, rhs)

  def bind(self, lhs, rhs):
    lhs_class = lhs.__class__
    if lhs_class is Var:
      self.bind_var(lhs.name, rhs)
    elif lhs_class is Tuple and rhs.__class__ is Tuple:
      assert len(lhs.elts) == len(rhs.elts)
      for lhs_elt, rhs_elt in zip(lhs.elts, rhs.elts):
        self.bind(lhs_elt, rhs_elt)

  def transform_lhs_Index(self, lhs):
    # lhs.value = self.transform_expr(lhs.value)
    lhs.index = self.transform_expr(lhs.index)
    if lhs.value.__class__ is Var:
      stored = self.bindings.get(lhs.value.name)
      if stored and stored.__class__ is Var:
        lhs.value = stored
    else:
      lhs.value = self.assign_temp(lhs.value, "array")
    return lhs

  def transform_lhs_Attribute(self, lhs):
    # lhs.value = self.transform_expr(lhs.value)
    return lhs

  def transform_ExprStmt(self, stmt):
    """Don't run an expression unless it possibly has a side effect"""

    v = self.transform_expr(stmt.value)
    if self.immutable(v):
      return None
    else:
      stmt.value = v
      return stmt

  def transform_Assign(self, stmt):
    lhs = stmt.lhs
    rhs = self.transform_expr(stmt.rhs)

    lhs_class = lhs.__class__
    rhs_class = rhs.__class__
    if lhs_class is Var:
      if rhs.type.__class__ is NoneT and self.use_counts.get(lhs.name,0) == 0:
        return self.transform_stmt(ExprStmt(rhs))
      else:
        self.bind_var(lhs.name, rhs)
        if rhs_class is not Var and \
           rhs_class is not Const and \
           self.immutable(rhs):
          self.available_expressions.setdefault(rhs, lhs)
    elif lhs_class is Tuple:
      self.bind(lhs, rhs)
    # assigning x[i] = x[i]
    # does nothing
    elif lhs_class is Index:
      if rhs_class is Index and \
         lhs.value == rhs.value and \
         lhs.index == rhs.index:
        # kill effect-free writes like x[i] = x[i]
        return None
      elif rhs_class is Var and \
           lhs.value.__class__ is Var and \
           lhs.value.name == rhs.name and \
           lhs.index.type.__class__ is TupleT and \
           all(elt_t == slice_none_t for elt_t in lhs.index.type.elt_types):
        # also kill x[:] = x
        return None
      else:
        lhs = self.transform_lhs_Index(lhs)
        # when assigning x[j] = [1,2,3]
        # just rewrite it as a sequence of element assignments 
        # to avoid 
        if lhs.type.__class__ is ArrayT and \
           lhs.type.rank == 1 and \
           rhs.__class__ is Array:
          lhs_slice = self.assign_temp(lhs, "lhs_slice")
          for (elt_idx, elt) in enumerate(rhs.elts):
            lhs_idx = self.index(lhs_slice, const_int(elt_idx), temp = False)
            self.assign(lhs_idx, elt)
          return None
        elif not self.is_simple(rhs):
          rhs = self.assign_temp(rhs)
    else:
      assert lhs_class is Attribute
      assert False, "Considering making attributes immutable"
      lhs = self.transform_lhs_Attribute(lhs)

    if rhs_class is Var and \
       rhs.name in self.bindings and \
       self.use_counts.get(rhs.name, 1) == 1:
      self.use_counts[rhs.name] = 0
      rhs = self.bindings[rhs.name]
    stmt.lhs = lhs
    
    stmt.rhs = rhs
    return stmt

  def transform_block(self, stmts):
    self.available_expressions.push()
    new_stmts = Transform.transform_block(self, stmts)
    self.available_expressions.pop()
    return new_stmts

  def transform_merge(self, phi_nodes, left_block, right_block):
    result = {}
    for (k, (left, right)) in phi_nodes.iteritems():
      new_left = self.transform_expr(left)
      new_right = self.transform_expr(right)

      if not isinstance(new_left, (Const, Var)):
        new_left = self.temp_in_block(new_left, left_block)
      if not isinstance(new_right, (Const, Var)):
        new_right = self.temp_in_block(new_right, right_block)

      if new_left == new_right:
        # if both control flows yield the same value then
        # we don't actually need the phi-bound variable, we can just
        # replace the left value everywhere
        self.set_binding(k, new_left)
      else:
        result[k] = new_left, new_right
    return result

  def transform_If(self, stmt):
    stmt.true = self.transform_block(stmt.true)
    stmt.false = self.transform_block(stmt.false)
    stmt.merge = self.transform_merge(stmt.merge,
                                      left_block = stmt.true,
                                      right_block = stmt.false)
    stmt.cond = self.transform_expr(stmt.cond)
    return stmt

  def transform_loop_condition(self, expr, outer_block, loop_body, merge):
    """Normalize loop conditions so they are just simple variables"""

    if self.is_simple(expr):
      return self.transform_expr(expr)
    else:
      loop_carried_vars = [name for name in collect_var_names(expr)
                           if name in merge]
      if len(loop_carried_vars) == 0:
        return expr

      left_values = [merge[name][0] for name in loop_carried_vars]
      right_values = [merge[name][1] for name in loop_carried_vars]

      left_cond = subst.subst_expr(expr, dict(zip(loop_carried_vars,
                                                  left_values)))
      if not self.is_simple(left_cond):
        left_cond = self.temp_in_block(left_cond, outer_block, name = "cond")

      right_cond = subst.subst_expr(expr, dict(zip(loop_carried_vars,
                                                   right_values)))
      if not self.is_simple(right_cond):
        right_cond = self.temp_in_block(right_cond, loop_body, name = "cond")

      cond_var = self.fresh_var(left_cond.type, "cond")
      merge[cond_var.name] = (left_cond, right_cond)
      return cond_var

    
  def transform_While(self, stmt):
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge(stmt.merge,
                                      left_block = self.blocks.current(),
                                      right_block = stmt.body)
    stmt.cond = \
        self.transform_loop_condition(stmt.cond,
                                      outer_block = self.blocks.current(),
                                      loop_body = stmt.body,
                                      merge = stmt.merge)
    return stmt

  def transform_ForLoop(self, stmt):
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge(stmt.merge,
                                      left_block = self.blocks.current(),
                                      right_block = stmt.body)
    stmt.start = self.transform_arg(stmt.start, 'start')
    stmt.stop = self.transform_arg(stmt.stop, 'stop')
    if self.is_none(stmt.step):
      stmt.step = syntax_helpers.one(stmt.start.type)
    else:
      stmt.step = self.transform_arg(stmt.step, 'step')


    # if a loop is only going to run for one iteration, might as well get rid of
    # it
    if stmt.start.__class__ is Const and \
       stmt.stop.__class__ is Const and \
       stmt.step.__class__ is Const:
      if stmt.start.value >= stmt.stop.value:
        for (var_name, (input_value, _)) in stmt.merge.iteritems():
          var = Var(var_name, input_value.type)
          self.blocks.append(Assign(var, input_value))
        return None
      elif stmt.start.value + stmt.step.value >= stmt.stop.value:
        for (var_name, (input_value, _)) in stmt.merge.iteritems():
          var = Var(var_name, input_value.type)
          self.blocks.append(Assign(var, input_value))
        self.assign(stmt.var, stmt.start)
        self.blocks.top().extend(stmt.body)
        return None
    return stmt

  def transform_Return(self, stmt):
    new_value = self.transform_expr(stmt.value)
    """
    if new_value.__class__ is Var and \
       new_value.name in self.use_counts and \
       self.use_counts[new_value.name] == 1 and \
       new_value.name in self.bindings:
      stored = self.bindings[stmt.value.name]
      if self.immutable(stored) and stored.__class__ is not AllPairs:
        print "Replacing %s => %s" % (stmt, stored)
        stmt.value = stored
        return stmt
    """
    if new_value != stmt.value:
      stmt.value = new_value
    return stmt



from core_types import Int32, Int64, Float32, Float64, ScalarT
from syntax import Var 
from transform import Transform 
from syntax_visitor import SyntaxVisitor  
from node import Node 
from syntax import Range, Const, ConstArray, ConstArrayLike, Array, Index

import syntax
import collect_vars
class AbstractValue(Node):
  _members = []

class ConstValue(AbstractValue):
  """
  All elements are the same constant 
  """
  _members = ['value', 'type']

zeros_float32 = ConstValue(0.0, Float32)
zeros_float64 = ConstValue(0.0, Float64)
zeros_int32 = ConstValue(0, Int32)
zeros_int64 = ConstValue(0, Int64)

class ConstElts(AbstractValue):
  """
  All elements known, represented as a numpy array 
  """ 
  _members = ['array']

class RangeArray(AbstractValue):
  """
  Result of a range expression 
  """
  _members = ['start', 'step', 'type']

class Unknown(AbstractValue):
  _members = []
  
unknown = Unknown()

class IndexElimAnalysis(SyntaxVisitor):
    def __init__(self):
      # map variables to (start, stop, step)  
      self.index_vars = {}
      
      # map array names to abstract values 
      self.array_values = {}
      
      # map each array name to a set of names which share the same 
      # memory 
      self.linked_arrays = {}
      
    def visit_Assign(self, stmt):
      lhs = stmt.lhs
      rhs = stmt.rhs 
      
      if lhs.__class__ is Var:
        lhs_name = stmt.lhs.name
        
        rhs_class = rhs.__class__ 
        if rhs_class is Var:
          linked_set = self.linked_arrays.get(lhs_name, set([lhs_name]))
          rhs_name = stmt.rhs.name
          linked_set.add(rhs_name)
          self.linked_arrays[lhs_name] = linked_set 
          self.linked_arrays[rhs_name] = linked_set
          if lhs_name in self.array_values:
            self.array_values[rhs_name] = self.array_values[lhs_name]
        elif rhs_class is Range:
          v = RangeArray(rhs.start, rhs.step, rhs.type.elt_type)
          self.array_values[lhs_name] = v 
      else:
        # if not a var, might be an index expression 
        for tainted_lhs_name in collect_vars.collect_var_names(lhs):
          for linked_name in self.linked_arrays.get(tainted_lhs_name, set([tainted_lhs_name])):
            self.array_values[linked_name] = unknown 
        
class IndexElim(Transform):
  def pre_apply(self, fn):
    analysis = IndexElimAnalysis() 
    analysis.visit_fn(fn)
    self.array_values = analysis.array_values
     
  def transform_Index(self, expr):

    if expr.value.__class__ is not Var: return expr 
    x = expr.value.name 
    if x not in self.array_values: return expr
    v = self.array_values[x]
    if v is unknown: return expr 
    if self.is_tuple(expr.index):
      indices = self.tuple_elts(expr.index)
    else:
      indices = [expr.index]
    if not all(isinstance(idx.type, ScalarT) for idx in indices): return expr
     
    n_indices = len(indices)
    if v.__class__ is ConstValue:
      if n_indices != v.type.rank: return expr
      return syntax.Const(v.value, v.type)
    elif v.__class__ is ConstElts  and all(idx.__class__ is syntax.Const for idx in indices):
      if n_indices != v.type.rank: return expr
      idx_values = [idx.value for idx in indices]
      return syntax.Const(v.array[tuple(idx_values)], type = v.type.elt_type)
    
    assert v.__class__ is RangeArray
    assert len(indices) == 1
    idx = indices[0]
    if idx.__class__ is syntax.Const and v.step.__class__ is Const and v.start.__class__ is Const:
      return syntax.Const(idx.value * v.step.value + v.start.value, type = v.type) 

    return self.add(v.start, self.mul(idx, v.step))
      


class AdverbSemantics(object):
  """
  Describe the behavior of adverbs in terms of lower-level value and iteration
  constructs.

  To get something other than an unfathomably slow interpreter, override all the
  methods of BaseSemantics and make them work for some other domain (such as
  types, shapes, or compiled expressions)
  """

  def build_slice_indices(self, rank, axis, idx):
    if rank == 1:
      assert axis == 0
      return idx

    indices = []
    for i in xrange(rank):
      if i == axis:
        indices.append(idx)
      else:
        s = self.slice_value(self.none, self.none, self.int(1))
        indices.append(s)
    return self.tuple(indices)

  def slice_along_axis(self, arr, axis, idx):
    r = self.rank(arr)
    if r > axis:
      index_tuple = self.build_slice_indices(r, axis, idx)
      return self.index(arr, index_tuple)
    else:
      return arr

  def delayed_elt(self, x, axis):
    return lambda idx: self.slice_along_axis(x, axis, idx)

  def delay_list(self, xs, axis):
    return [self.delayed_elt(x, axis) for x in xs]

  def force_list(self, delayed_elts, idx):
    return [e(idx) for e in delayed_elts]

  def sizes_along_axis(self, xs, axis):
    
    axis_sizes = [self.size_along_axis(x, axis)
                  for x in xs
                  if self.rank(x) > axis]

    assert len(axis_sizes) > 0
    # all arrays should agree in their dimensions along the
    # axis we're iterating over
    self.check_equal_sizes(axis_sizes)
    return axis_sizes

  def map_prelude(self, map_fn, xs, axis):
    axis_sizes = self.sizes_along_axis(xs, axis)
    return axis_sizes[0], self.delay_list(xs, axis)

  def acc_prelude(self, init, combine, delayed_map_result):
    zero = self.int(0)
    if init is None or self.is_none(init):
      return delayed_map_result(zero)
    else:
      # combine the provided initializer with
      # transformed first value of the data
      # in case we need to coerce up
      return self.invoke(combine, [init, delayed_map_result(zero)])

  def create_result(self, elt_type, inner_shape, outer_shape):
    if not self.is_tuple(outer_shape):
      outer_shape = self.tuple([outer_shape])
    result_shape = self.concat_tuples(outer_shape, inner_shape)
    result = self.alloc_array(elt_type, result_shape)
    return result

  def create_output_array(self, fn, inputs, extra_dims):
    if hasattr(self, "_create_output_array"):
      try:
        return self._create_output_array(fn, inputs, extra_dims)
      except:
        pass
  
    inner_result = self.invoke(fn, inputs)
    
    inner_shape = self.shape(inner_result)

    elt_t = self.elt_type(inner_result)
    res =  self.create_result(elt_t, inner_shape, extra_dims)
    return res 

  def output_slice(self, output, axis, idx):
    r = self.rank(output)
    if r > 1:
      output_indices = self.build_slice_indices(r, axis, idx)
    elif r == 1:
      output_idx = self.slice_value(idx, self.none, self.int(1))
      output_indices = self.tuple([output_idx])
    else:
      output_idx = self.slice_value(self.none, self.none, self.none)
      output_indices = self.tuple([output_idx])
    return self.index(output, output_indices)

  def eval_map(self, f, values, axis, output = None):
    niters, delayed_elts = self.map_prelude(f, values, axis)
    zero = self.int(0)
    first_elts = self.force_list(delayed_elts, zero)
    if output is None:
      output = self.create_output_array(f, first_elts, niters)
    def loop_body(idx):
      output_indices = self.build_slice_indices(self.rank(output), 0, idx)
      elt_result = self.invoke(f, [elt(idx) for elt in delayed_elts])
      self.setidx(output, output_indices, elt_result)
    self.loop(zero, niters, loop_body)
    return output

  def eval_reduce(self, map_fn, combine, init, values, axis):
    zero = self.int(0)
    one = self.int(1)
    
    if axis is  None or self.is_none(axis):
      assert len(values) == 1
      values = [self.ravel(values[0])]
      axis = 0
    
    niters, delayed_elts = self.map_prelude(map_fn, values, axis)
    first_acc_value = self.invoke(map_fn, [elt(zero) for elt in delayed_elts])
    if init is None or self.is_none(init):
      init = first_acc_value
    else:
      init = self.invoke(combine, [init, first_acc_value])
    def loop_body(acc, idx):
      elt = self.invoke(map_fn, [elt(idx) for elt in delayed_elts])
      new_acc_value = self.invoke(combine, [acc.get(), elt])
      acc.update(new_acc_value)
    return self.accumulate_loop(one, niters, loop_body, init)

  def eval_scan(self, map_fn, combine, emit, init, values, axis):

    niters, delayed_elts = self.map_prelude(map_fn, values, axis)
    def delayed_map_result(idx):
      return self.invoke(map_fn, self.force_list(delayed_elts, idx))
    init = self.acc_prelude(init, combine, delayed_map_result)
    output = self.create_output_array(emit, [init], niters)
    self.setidx(output, self.int(0), self.invoke(emit, [init]))
    def loop_body(acc, idx):
      output_indices = self.build_slice_indices(self.rank(output), 0, idx)
      new_acc_value = self.invoke(combine, [acc.get(), delayed_map_result(idx)])
      acc.update(new_acc_value)
      output_value = self.invoke(emit, [new_acc_value])
      self.setidx(output, output_indices, output_value)
    self.accumulate_loop(self.int(1), niters, loop_body, init)
    return output

  def eval_allpairs(self, fn, x, y, axis):
    if axis is None: 
      x = self.ravel(x)
      y = self.ravel(y)
      axis = 0
    nx = self.size_along_axis(x, axis)
    ny = self.size_along_axis(y, axis)
    outer_shape = self.tuple( [nx, ny] )
    zero = self.int(0)
    first_x = self.slice_along_axis(x, axis, zero)
    first_y = self.slice_along_axis(y, axis, zero)
    output =  self.create_output_array(fn, [first_x, first_y], outer_shape)
    def outer_loop_body(i):
      xi = self.slice_along_axis(x, axis, i)
      def inner_loop_body(j):
        yj = self.slice_along_axis(y, axis, j)
        out_idx = self.tuple([i,j])
        self.setidx(output, out_idx, self.invoke(fn, [xi, yj]))
      self.loop(zero, ny, inner_loop_body)
    self.loop(zero, nx, outer_loop_body)
    return output
  
  def eval_index_map(self, fn, shape, output = None): 
    dims = self.tuple_elts(shape)
    if len(dims) == 1:
      shape = dims[0]
      
    if output is None:
      output = self.create_output_array(fn, [shape], shape)

    n_loops = len(dims)
    def build_loops(index_vars = ()):
      n_indices = len(index_vars)
      if n_indices == n_loops:
        if n_indices > 1:
          idx_tuple = self.tuple(index_vars)
        else:
          idx_tuple = index_vars[0]
        elt_result =  self.invoke(fn, (idx_tuple,))
        self.setidx(output, index_vars, elt_result)
      else:
        def loop_body(idx):
          build_loops(index_vars + (idx,))
        self.loop(self.int(0), dims[n_indices], loop_body)
    build_loops()
    return output 
    
    """ 
    def loop_body(idx):
      output_indices = self.build_slice_indices(self.rank(output), 0, idx)
      elt_result = self.invoke(f, [elt(idx) for elt in delayed_elts])
      self.setidx(output, output_indices, elt_result)
    self.loop(zero, niters, loop_body)
    return output
    """
  def eval_index_reduce(self, fn, combine, shape, init = None):

    dims = self.tuple_elts(shape)
    n_loops = len(dims)
    
    zero = self.int(0)
    
    if init is not None or not self.is_none(init):
      if n_loops > 1:
        zeros = self.tuple([zero for _ in xrange(n_loops)])
        init = self.invoke(fn, [zeros])
      else:
        init = self.invoke(fn, [zero])
        
    def build_loops(index_vars, acc):
    
      n_indices = len(index_vars)
      if n_indices > 0:
        acc_value = acc.get()
      else:
        acc_value = acc 
      if n_indices == n_loops:
        
        idx_tuple = self.tuple(index_vars) if n_indices > 1 else index_vars[0] 
        elt_result =  self.invoke(fn, (idx_tuple,))
        acc.update(self.invoke(combine, (acc_value, elt_result)))
        return acc.get()
      
      def loop_body(acc, idx):
        new_value = build_loops(index_vars + (idx,), acc = acc)
        acc.update(new_value)
        return new_value
      return self.accumulate_loop(self.int(0), dims[n_indices], loop_body, acc_value)
    return build_loops(index_vars = (), acc = init)
   
    
    
    


import escape_analysis

from array_type import ArrayT
from collect_vars import collect_var_names
from core_types import ScalarT
from find_local_arrays import FindLocalArrays
from syntax import AllocArray, ArrayView, Index, Struct, Var
from syntax import Adverb
from transform import Transform
from usedef import UseDefAnalysis

class CopyElimination(Transform):
  def apply(self, fn):
    if all(isinstance(t, ScalarT) for t in fn.type_env.itervalues()):
      return fn
    else:
      return Transform.apply(self, fn)

  def pre_apply(self, fn):
    find_local_arrays = FindLocalArrays()
    find_local_arrays.visit_fn(fn)

    self.local_alloc = find_local_arrays.local_allocs
    self.local_arrays = find_local_arrays.local_arrays

    escape_info = escape_analysis.run(fn)
    self.may_escape = escape_info.may_escape
    self.may_alias = escape_info.may_alias

    self.usedef = UseDefAnalysis()
    self.usedef.visit_fn(fn)

    self.pointers_by_size = {}
    self.arrays_by_size = {}

  def no_array_aliases(self, array_name):
    alias_set = self.may_alias.get(array_name, [])
    array_aliases = [name for name in alias_set
                     if self.type_env[name].__class__ is ArrayT]
    # you're allowed one alias for yourself, but
    # any extras are other arrays with whom you share data
    # BEWARE: this will get convoluted and probably broken
    # if we ever have mutable compound objects in arrays
    return len(array_aliases) <= 1

  def is_array_alloc(self, expr):
    return expr.__class__ in [ArrayView, Struct, AllocArray] or isinstance(expr, Adverb)
    
  def transform_Assign(self, stmt):
    # pattern match only on statements of the form
    # dest[complex_indexing] = src
    # when:
    #   1) dest hasn't been used before as a value
    #   2) src doesn't escape
    #   3) src was locally allocated
    # ...then transform the code so instead of allocating src

    if stmt.lhs.__class__ is Index and  stmt.lhs.value.__class__ is Var:
      lhs_name = stmt.lhs.value.name

      if lhs_name not in self.usedef.first_use and \
         lhs_name not in self.may_escape and \
         self.no_array_aliases(lhs_name):
        # why assign to an array if it never gets used?
        return None
      elif stmt.lhs.type.__class__ is ArrayT and stmt.rhs.__class__ is Var:
        curr_path = self.usedef.stmt_paths[id(stmt)]
        rhs_name = stmt.rhs.name
        if lhs_name not in self.usedef.first_use or \
           self.usedef.first_use[lhs_name] > curr_path:
          if self.usedef.last_use[rhs_name] == curr_path and \
             rhs_name not in self.may_escape and \
             rhs_name in self.local_arrays:
            array_stmt = self.local_arrays[rhs_name]
            prev_path = self.usedef.stmt_paths[id(array_stmt)]
            if self.is_array_alloc(array_stmt.rhs) and \
               all(self.usedef.created_on[lhs_depends_on] < prev_path
                   for lhs_depends_on in collect_var_names(stmt.lhs)):
              array_stmt.rhs = stmt.lhs
              return None
    return stmt



import numpy as np

import parakeet 
import testing_helpers

def erode(x, shape):
  return parakeet.pmap2(min, x, shape)

x = np.array([[1,2,3],[4,5,6]])

def test_erode_identity_shape():
  testing_helpers.expect(erode, [x, (1,1)], x)

def test_erode():
  shape = (3,3)
  expected = np.array([[1, 1, 2], [1,1,2]])
  testing_helpers.expect(erode, [x, shape], expected)

if __name__ == '__main__':
  testing_helpers.run_local_tests()



import __builtin__
import ast
import inspect
import types

import numpy as np

import config
import core_types 
import lib 
import names
import nested_blocks 
import prims
import scoped_dict
import syntax
import syntax_helpers 

from args import FormalArgs, ActualArgs
from collections import OrderedDict
from function_registry import already_registered_python_fn
from function_registry import register_python_fn, lookup_python_fn
from decorators import macro, jit 
from names import NameNotFound
from prims import Prim, prim_wrapper
from python_ref import GlobalValueRef, GlobalNameRef, ClosureCellRef

from subst import subst_expr, subst_stmt_list
from syntax import Assign, If, ForLoop, Var, PrimCall, Map
from syntax_helpers import none, true, false, one_i64, zero_i64

from mappings import function_mappings, method_mappings, property_mappings 

class UnsupportedSyntax(Exception):
  def __init__(self, node, function_name = None, filename = None):
    self.function_name = function_name 
    self.filename = filename 
    self.node = node
    
  def __str__(self):
    if self.function_name is not None and \
       self.filename is not None and \
       self.node.lineno is not None:
      return "Parakeet doesn't support %s from function '%s' in file %s on line %d" % \
        (self.node.__class__.__name__, self.function_name, self.filename, self.node.lineno)
    elif self.function_name is not None:
      return "Parakeet doesn't support %s in '%s'" % \
        (self.node.__class__.__name__, self.function_name)
    else:
      return "Parakeet doesn't support %s" % self.node.__class__.__name__
  
class ExternalValue(object):
  """
  Wrap up references global values with this class
  """ 
  def __init__(self, python_value):
    self.value = python_value
    
  def __str__(self):
    return "ExternalValue(%s)" % self.value


def mk_reduce_call(fn, positional, init = None):
  init = syntax_helpers.none if init is None else init
  axis = syntax_helpers.zero_i64
  return syntax.Reduce(fn = translate_function_value(lib.identity), 
                       combine = fn, 
                       args = positional, 
                       axis = axis, 
                       init = init)

def mk_simple_fn(mk_body, input_name = "x", fn_name = "cast"):
  unique_arg_name = names.fresh(input_name)
  unique_fn_name = names.fresh(fn_name)
  var = syntax.Var(unique_arg_name)
  formals = FormalArgs()
  formals.add_positional(unique_arg_name, input_name)
  body = mk_body(var)
  return syntax.Fn(unique_fn_name, formals, body)
  
_function_wrapper_cache = {}
def mk_wrapper_function(p):
  """
  Generate wrappers for builtins and primitives
  """
  if p in _function_wrapper_cache:
    return _function_wrapper_cache[p]
  if isinstance(p, prims.Prim):
    f = prims.prim_wrapper(p)
  elif p in prims.prim_lookup_by_value:
    f = prims.prim_wrapper(prims.prim_lookup_by_value[p]) 
  else:
    assert isinstance(p, types.BuiltinFunctionType)
    assert p.__name__ in lib.__dict__, "Unsupported builtin: %s" % (p,)
    f = translate_function_value(lib.__dict__[p.__name__]) 
  _function_wrapper_cache[p] = f
  return f
  
       
def is_hashable(x):
  try:
    hash(x)
    return True
  except:
    return False 

def is_prim(v):
  return isinstance(v, (Prim)) or (is_hashable(v) and v in prims.prim_lookup_by_value) 

def is_builtin_function(v):
  return isinstance(v, (types.TypeType, types.BuiltinFunctionType))
  
def is_user_function(v):
  return isinstance(v, (types.FunctionType, jit, macro))

def is_function_value(v):
  return is_user_function(v) or is_builtin_function(v) or is_prim(v) 
    
def is_static_value(v):
  return syntax_helpers.is_python_constant(v) or \
         type(v) is np.dtype  or \
         is_function_value(v)

def value_to_syntax(v):
  if syntax_helpers.is_python_constant(v):
    return syntax_helpers.const(v)
  elif isinstance(v, np.dtype):
    x = names.fresh("x")
    fn_name = names.fresh("cast") 
    formals = FormalArgs()
    formals.add_positional(x, "x")
    body = [syntax.Return(syntax.Cast(syntax.Var(x), type=core_types.from_dtype(v)))]
    return syntax.Fn(fn_name, formals, body)
  else:
    assert is_function_value(v), "Can't make value %s : %s into static syntax" % (v, type(v))
    return translate_function_value(v)  
    
class AST_Translator(ast.NodeVisitor):
  def __init__(self, 
               globals_dict=None, 
               closure_cell_dict=None,
               parent=None, 
               function_name = None, 
               filename = None):
    # assignments which need to get prepended at the beginning of the
    # function
    self.globals = globals_dict
    self.blocks = nested_blocks.NestedBlocks()

    self.parent = parent 
    self.scopes = scoped_dict.ScopedDictionary()
    
    self.globals_dict = globals_dict 
    self.closure_cell_dict = closure_cell_dict 
    
    # mapping from names/paths to either a closure cell reference or a 
    # global value 
    self.python_refs = OrderedDict()
    
    self.original_outer_names = []
    self.localized_outer_names = []
    self.filename = filename 
    self.function_name = function_name 
    self.push()

  def push(self, scope = None, block = None):
    if scope is None:
      scope = {}
    if block is None:
      block = []
    self.scopes.push(scope)
    self.blocks.push(block)


  def pop(self):
    scope = self.scopes.pop()
    block = self.blocks.pop()
    return scope, block
  
  def fresh_name(self, original_name):
    fresh_name = names.fresh(original_name)
    self.scopes[original_name] = fresh_name
    return fresh_name
    
  def fresh_names(self, original_names):
    return map(self.fresh_name, original_names)

  def fresh_var(self, name):
    return syntax.Var(self.fresh_name(name))

  def fresh_vars(self, original_names):
    return map(self.fresh_var, original_names)
  
  def current_block(self):
    return self.blocks.top()

  def current_scope(self):
    return self.scopes.top()


  def syntax_to_value(self, expr):
    if isinstance(expr, ast.Num):
      return expr.n
    elif isinstance(expr, ast.Tuple):
      return tuple(self.syntax_to_value(elt) for elt in expr.elts)
    elif isinstance(expr, ast.Name):
      return self.lookup_global(expr.id)
    elif isinstance(expr, ast.Attribute):
      left = self.syntax_to_value(expr.value)
      if isinstance(left, ExternalValue):
        left = left.value 
      return getattr(left, expr.attr) 

  def lookup_global(self, key):
    if isinstance(key, (list, tuple)):
      assert len(key) == 1
      key = key[0]
    else:
      assert isinstance(key, str), "Invalid global key: %s" % (key,)

    if self.globals:
      if key in self.globals:
        return self.globals[key]
      elif key in __builtin__.__dict__:

        return __builtin__.__dict__[key]
      else:
        assert False, "Couldn't find global name %s" % key
    else:
      assert self.parent is not None
      return self.parent.lookup_global(key)
    
  def is_global(self, key):
    if isinstance(key, (list, tuple)):
      key = key[0]
    if key in self.scopes:
      return False
    elif self.closure_cell_dict and key in self.closure_cell_dict:
      return False
    if self.globals:
      return key in self.globals or key in __builtins__  
    assert self.parent is not None 
    return self.parent.is_global(key)
  
  
  def lookup(self, name):
    #if name in reserved_names:
    #  return reserved_names[name]
    if name in self.scopes:
      return Var(self.scopes[name])
    elif self.parent:
      # don't actually keep the outer binding name, we just
      # need to check that it's possible and tell the outer scope
      # to register any necessary python refs
      local_name = names.fresh(name)
      self.scopes[name] = local_name
      self.original_outer_names.append(name)
      self.localized_outer_names.append(local_name)
      return Var(local_name)
    elif self.closure_cell_dict and name in self.closure_cell_dict:
      ref = ClosureCellRef(self.closure_cell_dict[name], name)
      for (local_name, other_ref) in self.python_refs.iteritems():
        if ref == other_ref:
          return Var(local_name)
      local_name = names.fresh(name)
      self.scopes[name] = local_name
      self.original_outer_names.append(name)
      self.localized_outer_names.append(local_name)
      self.python_refs[local_name] = ref
      return Var(local_name)
    elif self.is_global(name):
      value = self.lookup_global(name)
      if is_static_value(value): 
        return value_to_syntax(value)
      else:
        return ExternalValue(value)
      
        
      #else:
      #  assert False, "External values must be scalars or functions"
    else:
      raise NameNotFound(name)
      
      
  def visit_list(self, nodes):
    return map(self.visit, nodes)


  def tuple_arg_assignments(self, elts, var):
    """
    Recursively decompose a nested tuple argument like
      def f((x,(y,z))):
        ...
    into a single name and a series of assignments:
      def f(tuple_arg):
        x = tuple_arg[0]
        tuple_arg_elt = tuple_arg[1]
        y = tuple_arg_elt[0]
        z = tuple_arg_elt[1]
    """

    assignments = []
    for (i, sub_arg) in enumerate(elts):
      if isinstance(sub_arg, ast.Tuple):
        name = "tuple_arg_elt"
      else:
        assert isinstance(sub_arg, ast.Name)
        name = sub_arg.id
      lhs = self.fresh_var(name)
      stmt = Assign(lhs, syntax.Index(var, syntax.Const(i)))
      assignments.append(stmt)
      if isinstance(sub_arg, ast.Tuple):
        more_stmts = self.tuple_arg_assignments(sub_arg.elts, lhs)
        assignments.extend(more_stmts)
    return assignments

  def translate_args(self, args):
    assert not args.kwarg
    formals = FormalArgs()
    assignments = []
    for arg in args.args:
      if isinstance(arg, ast.Name):
        visible_name = arg.id
        local_name = self.fresh_name(visible_name)
        formals.add_positional(local_name, visible_name)
      else:
        assert isinstance(arg, ast.Tuple)
        arg_name = self.fresh_name("tuple_arg")
        formals.add_positional(arg_name)
        var = Var(arg_name)
        stmts = self.tuple_arg_assignments(arg.elts, var)
        assignments.extend(stmts)

    n_defaults = len(args.defaults)
    if n_defaults > 0:
      local_names = formals.positional[-n_defaults:]
      for (k,expr) in zip(local_names, args.defaults):
        formals.defaults[k] = self.syntax_to_value(expr)


    if args.vararg:
      assert isinstance(args.vararg, str)
      formals.starargs = self.fresh_name(args.vararg)

    return formals, assignments


  def visit_Name(self, expr):
    assert isinstance(expr, ast.Name), "Expected AST Name object: %s" % expr
    return self.lookup(expr.id)

  def create_phi_nodes(self, left_scope, right_scope, new_names = {}):
    """
    Phi nodes make explicit the possible sources of each variable's values and
    are needed when either two branches merge or when one was optionally taken.
    """
    merge = {}
    for (name, ssa_name) in left_scope.iteritems():
      left = Var(ssa_name)
      if name in right_scope:
        right = Var(right_scope[name])
      else:
        right = self.lookup(name)

      if name in new_names:
        new_name = new_names[name]
      else:
        new_name = self.fresh_name(name)
      merge[new_name] = (left, right)

    for (name, ssa_name) in right_scope.iteritems():
      if name not in left_scope:
        try:
          left = self.lookup(name)
          right = Var(ssa_name)

          if name in new_names:
            new_name = new_names[name]
          else:
            new_name = self.fresh_name(name)
          merge[new_name] = (left, right)
        except names.NameNotFound:
          # for now skip over variables which weren't defined before
          # a control flow split, which means that loop-local variables
          # can't be used after the loop.
          # TODO: Fix this. Maybe with 'undef' nodes?
          pass
    return merge


  def visit_Index(self, expr):
    return self.visit(expr.value)

  def visit_Ellipsis(self, expr):
    raise RuntimeError("Ellipsis operator unsupported")

  def visit_Slice(self, expr):
    """
    x[l:u:s]
    Optional fields
      expr.lower
      expr.upper
      expr.step
    """

    start = self.visit(expr.lower) if expr.lower else none
    stop = self.visit(expr.upper) if expr.upper else none
    step = self.visit(expr.step) if expr.step else none
    return syntax.Slice(start, stop, step)

  def visit_ExtSlice(self, expr):
    slice_elts = map(self.visit, expr.dims)
    if len(slice_elts) > 1:
      return syntax.Tuple(slice_elts)
    else:
      return slice_elts[0]
    

  def visit_UnaryOp(self, expr):
    ssa_val = self.visit(expr.operand)
    prim = prims.find_ast_op(expr.op)
    return syntax.PrimCall(prim, [ssa_val])

  def visit_BinOp(self, expr):
    ssa_left = self.visit(expr.left)
    ssa_right = self.visit(expr.right)
    prim = prims.find_ast_op(expr.op)
    return syntax.PrimCall(prim, [ssa_left, ssa_right])

  def visit_BoolOp(self, expr):
    values = map(self.visit, expr.values)
    prim = prims.find_ast_op(expr.op)
    # Python, strangely, allows more than two arguments to
    # Boolean operators
    result = values[0]
    for v in values[1:]:
      result = syntax.PrimCall(prim, [result, v])
    return result

  def visit_Compare(self, expr):
    lhs = self.visit(expr.left)
    assert len(expr.ops) == 1
    prim = prims.find_ast_op(expr.ops[0])
    assert len(expr.comparators) == 1
    rhs = self.visit(expr.comparators[0])
    return syntax.PrimCall(prim, [lhs, rhs])

  def visit_Subscript(self, expr):
    value = self.visit(expr.value)
    index = self.visit(expr.slice)
    return syntax.Index(value, index)

  def generic_visit(self, expr):
    raise UnsupportedSyntax(expr, 
                            function_name = self.function_name, 
                            filename = self.filename)
  
  def translate_builtin_call(self, value, positional, keywords_dict):
    if value is sum:
      return mk_reduce_call(prims.prim_wrapper(prims.add), positional, zero_i64)
    elif value is max:
      if len(positional) == 1:
        return mk_reduce_call(prims.prim_wrapper(prims.maximum), positional)
      else:
        assert len(positional) == 2
        return syntax.PrimCall(prims.maximum, positional)
    elif value is min:
      if len(positional) == 1:
        return mk_reduce_call(prims.prim_wrapper(prims.minimum), positional)
      else:
        assert len(positional) == 2
        return syntax.PrimCall(prims.minimum, positional)
    elif value is map:
      assert len(keywords_dict) == 0
      assert len(positional) > 1
      axis = keywords_dict.get("axis", None)
      return Map(fn = positional[0], args = positional[1:], axis = axis)
    elif isinstance(value, (types.BuiltinFunctionType, types.TypeType)) and \
         value.__name__ in lib.__dict__:
      parakeet_equiv = lib.__dict__[value.__name__]
      if isinstance(parakeet_equiv, macro):
        return parakeet_equiv.transform(positional, keywords_dict)
    fn = value_to_syntax(value)
    return syntax.Call(fn, ActualArgs(positional, keywords_dict))
    
      
  def visit(self, node):
    res = ast.NodeVisitor.visit(self, node)
    return res 
    
  def translate_value_call(self, value, positional, keywords_dict= {}, starargs_expr = None):
    if value in function_mappings:
      value = function_mappings[value]
      
    if isinstance(value, macro):
      return value.transform(positional, keywords_dict)
    elif is_user_function(value):
      return syntax.Call(translate_function_value(value), 
                           ActualArgs(positional, keywords_dict, starargs_expr))
    else:
      return self.translate_builtin_call(value, positional, keywords_dict)
    
  def visit_Call(self, expr):
    """
    TODO: 
    The logic here is broken and haphazard, eventually try to handle nested
    scopes correctly, along with globals, cell refs, etc..
    """
    
    fn, args, keywords_list, starargs, kwargs = \
        expr.func, expr.args, expr.keywords, expr.starargs, expr.kwargs
    assert kwargs is None, "Dictionary of keyword args not supported"

    positional = self.visit_list(args)

    keywords_dict = {}
    for kwd in keywords_list:
      keywords_dict[kwd.arg] = self.visit(kwd.value)

    if starargs:
      starargs_expr = self.visit(starargs)
    else:
      starargs_expr = None
    
    def is_attr_chain(expr):
      return isinstance(expr, ast.Name) or \
             (isinstance(expr, ast.Attribute) and is_attr_chain(expr.value))
    def extract_attr_chain(expr):
      if isinstance(expr, ast.Name):
        return [expr.id]
      else:
        base = extract_attr_chain(expr.value)
        base.append(expr.attr)
        return base
    
    def lookup_attr_chain(names):
      value = self.lookup_global(names[0])
    
      for name in names[1:]:
        if hasattr(value, name):
          value = getattr(value, name)
        else:
          try:
            value = value[name]
          except:
            assert False, "Couldn't find global name %s" % ('.'.join(names))
      return value
          
    if is_attr_chain(fn):
      names = extract_attr_chain(fn)
      
      if self.is_global(names):
        return self.translate_value_call(lookup_attr_chain(names), 
                                         positional, keywords_dict, starargs_expr)
    fn_node = self.visit(fn)    
    if isinstance(fn_node, syntax.Expr):
      actuals = ActualArgs(positional, keywords_dict, starargs_expr)
      return syntax.Call(fn_node, actuals)
    else:
      assert isinstance(fn_node, ExternalValue)
      return self.translate_value_call(fn_node.value, 
                                       positional, keywords_dict, starargs_expr)

  def visit_List(self, expr):
    return syntax.Array(self.visit_list(expr.elts))
    
  def visit_Expr(self, expr):
    # dummy assignment to allow for side effects on RHS
    lhs = self.fresh_var("dummy")
    if isinstance(expr.value, ast.Str):
      return syntax.Assign(lhs, zero_i64)
      # return syntax.Comment(expr.value.s.strip().replace('\n', ''))
    else:
      rhs = self.visit(expr.value)
      return syntax.Assign(lhs, rhs)

    
  def visit_GeneratorExp(self, expr):
    return self.visit_ListComp(expr)
    
  def visit_ListComp(self, expr):
    gens = expr.generators
    assert len(gens) == 1
    gen = gens[0]
    target = gen.target
    if target.__class__ is ast.Name:
      arg_vars = [target]
    else:
      assert target.__class__ is ast.Tuple and all(e.__class__ is ast.Name for e in target.elts),\
       "Expected comprehension target to be variable or tuple of variables, got %s" % ast.dump(target)
      arg_vars = [ast.Tuple(elts = target.elts)]
    # build a lambda as a Python ast representing 
    # what we do to each element 

    args = ast.arguments(args = arg_vars, 
                         vararg = None,  
                         kwarg = None,  
                         defaults = ())

    fn = translate_function_ast(name = "comprehension_map", 
                                args = args, 
                                body = [ast.Return(expr.elt)], 
                                parent = self)

    seq = self.visit(gen.iter)
    ifs = gen.ifs
    assert len(ifs) == 0, "Parakeet: Conditions in array comprehensions not yet supported"
    return Map(fn = fn, args=(seq,), axis = zero_i64)
      
  def visit_Attribute(self, expr):
    # TODO:
    # Recursive lookup to see if:
    #  (1) base object is local, if so-- create chain of attributes
    #  (2) base object is global but an adverb primitive-- use it locally
    #      without adding it to nonlocals
    #  (3) not local at all-- in which case, add the whole chain of strings
    #      to nonlocals
    #
    #  AN IDEA:
    #     Allow external values to be brought into the syntax tree as 
    #     a designated ExternalValue node
    #     and then here check if the LHS is an ExternalValue and if so, 
    #     pull out the value. If it's a constant, then make it into syntax, 
    #     if it's a function, then parse it, else raise an error. 
    #
    value = self.visit(expr.value)
    attr = expr.attr
    if isinstance(value, ExternalValue):
      value = value.value 
      value = getattr(value, attr)
      if is_static_value(value):
        return value_to_syntax(value)
      else:
        return ExternalValue(value) 
    elif attr in property_mappings:
      fn = property_mappings[attr]
      if isinstance(fn, macro):
        return fn.transform( [value] )
      else:
        return syntax.Call(translate_function_value(fn),
                            ActualArgs(positional = (value,)))  
    elif attr in method_mappings:
      fn_python = method_mappings[attr]
      fn_syntax = translate_function_value(fn_python)
      return syntax.Closure(fn_syntax, args=(value,))
    else:
      assert False, "Attribute %s not supported" % attr 
    # return syntax.Attribute(value, expr.attr)

  def visit_Num(self, expr):
    return syntax.Const(expr.n)

  def visit_Tuple(self, expr):
    return syntax.Tuple(self.visit_list(expr.elts))

  def visit_IfExp(self, expr):
    temp1, temp2, result = self.fresh_vars(["if_true", "if_false", "if_result"])
    cond = self.visit(expr.test)
    true_block = [Assign(temp1, self.visit(expr.body))]
    false_block = [Assign(temp2, self.visit(expr.orelse))]
    merge = {result.name : (temp1, temp2)}
    if_stmt = If(cond, true_block, false_block, merge)
    self.current_block().append(if_stmt)
    return result

  def visit_lhs(self, lhs):
    if isinstance(lhs, ast.Name):
      return self.fresh_var(lhs.id)
    elif isinstance(lhs, ast.Tuple):
      return syntax.Tuple( map(self.visit_lhs, lhs.elts))
    else:
      # in case of slicing or attributes
      res = self.visit(lhs)
      return res

  def visit_Assign(self, stmt):
    # important to evaluate RHS before LHS for statements like 'x = x + 1'
    ssa_rhs = self.visit(stmt.value)
    ssa_lhs = self.visit_lhs(stmt.targets[0])
    return Assign(ssa_lhs, ssa_rhs)
  
  def visit_AugAssign(self, stmt):
    ssa_incr = self.visit(stmt.value)
    ssa_old_value = self.visit(stmt.target)
    ssa_new_value = self.visit_lhs(stmt.target)
    prim = prims.find_ast_op(stmt.op) 
    return Assign(ssa_new_value, PrimCall(prim, [ssa_old_value, ssa_incr]))

  def visit_Return(self, stmt):
    return syntax.Return(self.visit(stmt.value))

  def visit_If(self, stmt):
    cond = self.visit(stmt.test)
    true_scope, true_block  = self.visit_block(stmt.body)
    false_scope, false_block = self.visit_block(stmt.orelse)
    merge = self.create_phi_nodes(true_scope, false_scope)
    return syntax.If(cond, true_block, false_block, merge)

  def visit_loop_body(self, body, *exprs):
    merge = {}
    substitutions = {}
    curr_scope = self.current_scope()
    exprs = [self.visit(expr) for expr in exprs]
    scope_after, body = self.visit_block(body)
    for (k, name_after) in scope_after.iteritems():
      if k in self.scopes:
        name_before = self.scopes[k]
        new_name = names.fresh(k + "_loop")
        merge[new_name] = (Var(name_before), Var(name_after))
        substitutions[name_before]  = new_name
        curr_scope[k] = new_name
    
    exprs = [subst_expr(expr, substitutions) for expr in exprs]
    body = subst_stmt_list(body, substitutions)
    return body, merge, exprs 

  def visit_While(self, stmt):
    assert not stmt.orelse
    body, merge, (cond,) = self.visit_loop_body(stmt.body, stmt.test)
    return syntax.While(cond, body, merge)

  def visit_For(self, stmt):
    assert not stmt.orelse 
    var = self.visit_lhs(stmt.target)
    assert isinstance(var, Var)
    seq = self.visit(stmt.iter)
    body, merge, _ = self.visit_loop_body(stmt.body)
    if isinstance(seq, syntax.Range):
      return ForLoop(var, seq.start, seq.stop, seq.step, body, merge)
    else:
      seq_name = self.fresh_name("seq")
      seq_var = Var(seq_name)
      self.current_block().append(Assign(seq_var, seq))
      len_fn = translate_function_value(len)
      n = syntax.Call(len_fn, ActualArgs([seq_var]))
      start = zero_i64
      step = one_i64
      loop_counter_name = self.fresh_name('loop_counter')
      loop_var = Var(loop_counter_name)
      body = [Assign(var, syntax.Index(seq_var, loop_var))] + body
      return ForLoop(loop_var, start, n, step, body, merge)
    
  def visit_block(self, stmts):
    self.push()
    curr_block = self.current_block()
    for stmt in stmts:
      parakeet_stmt = self.visit(stmt)
      curr_block.append(parakeet_stmt)
    return self.pop()

  def visit_FunctionDef(self, node):
    """
    Translate a nested function
    """
    fundef = translate_function_ast(node.name, node.args, node.body, parent = self)
    local_var = self.fresh_var(node.name)
    return Assign(local_var, fundef)

  def visit_Lambda(self, node):
    return translate_function_ast("lambda", node.args, [ast.Return(node.body)], parent = self)
    
def translate_function_ast(name, 
                           args, 
                           body, 
                           globals_dict = None,
                           closure_vars = [], 
                           closure_cells = [],
                           parent = None, 
                           filename = None):
  """
  Helper to launch translation of a python function's AST, and then construct
  an untyped parakeet function from the arguments, refs, and translated body.
  """
  
  assert len(closure_vars) == len(closure_cells)
  closure_cell_dict = dict(zip(closure_vars, closure_cells))
  
  if filename is None and parent is not None:
    filename = parent.filename 
    
  translator = AST_Translator(globals_dict, closure_cell_dict, 
                              parent, function_name = name, filename = filename)

  ssa_args, assignments = translator.translate_args(args)
  _, body = translator.visit_block(body)
  body = assignments + body

  ssa_fn_name = names.fresh(name)

  # if function was nested in parakeet, it can have references to its
  # surrounding parakeet scope, which can't be captured with a python ref cell
  original_outer_names = translator.original_outer_names
  localized_outer_names = translator.localized_outer_names
  python_refs = translator.python_refs 
  ssa_args.prepend_nonlocal_args(localized_outer_names)
  if globals_dict:
    assert parent is None
    assert len(original_outer_names) == len(python_refs)
    return syntax.Fn(ssa_fn_name, ssa_args, body, python_refs.values(), [])
  else:
    assert parent
    fn = syntax.Fn(ssa_fn_name, ssa_args, body, [], original_outer_names)
    if len(original_outer_names) > 0:
      outer_ssa_vars = [parent.lookup(x) for x in original_outer_names]
      return syntax.Closure(fn, outer_ssa_vars)
    else:
      return fn

def strip_leading_whitespace(source):
  lines = source.splitlines()
  assert len(lines) > 0
  first_line = lines[0]
  n_removed = len(first_line) - len(first_line.lstrip())
  if n_removed > 0:
    return '\n'.join(line[n_removed:] for line in lines)
  else:
    return source

def translate_function_source(source, 
                              globals_dict, 
                              closure_vars = [],
                              closure_cells = [],
                              filename = None):
  assert len(closure_vars) == len(closure_cells)
  syntax = ast.parse(strip_leading_whitespace(source))

  if isinstance(syntax, (ast.Module, ast.Interactive)):
    assert len(syntax.body) == 1
    syntax = syntax.body[0]
  elif isinstance(syntax, ast.Expression):
    syntax = syntax.body
  
  assert isinstance(syntax, ast.FunctionDef), \
      "Unexpected Python syntax node: %s" % ast.dump(syntax)
  return translate_function_ast(syntax.name, 
                                syntax.args, 
                                syntax.body, 
                                globals_dict, 
                                closure_vars,
                                closure_cells, 
                                filename = filename)

def translate_function_value(fn, _currently_processing = set([])):
  if fn in function_mappings:
    fn = function_mappings[fn]
  
  if fn in prims.prim_lookup_by_value:
    fn = prims.prim_lookup_by_value[fn]
  
  # if the function has been wrapped with a decorator, unwrap it 
  while isinstance(fn, jit):
    fn = fn.f 
  
  # print ">>", fn   
  assert type(fn) not in (types.BuiltinFunctionType, types.TypeType), \
    "Unsupported primitive: %s" % (fn,) 

  if already_registered_python_fn(fn):
    return lookup_python_fn(fn)
  
  assert is_hashable(fn), "Can't convert unhashable value: %s" % (fn,)
  assert fn not in _currently_processing, \
    "Recursion detected through function value %s" % (fn,)
  _currently_processing.add(fn)
  original_fn = fn
  
  if isinstance(fn, Prim):
    fundef = prim_wrapper(fn)
  
    
  # TODO: Right now we can only deal with adverbs over a fixed axis and fixed
  # number of args due to lack of support for a few language constructs:
  # - variable number of args packed as a tuple i.e. *args
  # - keyword arguments packed as a...? ...struct of some kind? i.e. **kwds
  # - unpacking tuples and unpacking structs

  elif isinstance(fn, macro):
    fundef = fn.as_fn()
  else:
    assert hasattr(fn, 'func_globals'), "Expected function to have globals: %s" % fn
    assert hasattr(fn, 'func_closure'), "Expected function to have closure cells: %s" % fn
    assert hasattr(fn, 'func_code'), "Expected function to have code object: %s" % fn
    source = inspect.getsource(fn)
    filename = inspect.getsourcefile(fn)
    globals_dict = fn.func_globals

    free_vars = fn.func_code.co_freevars
    closure_cells = fn.func_closure
    if closure_cells is None: closure_cells = ()
    fundef = translate_function_source(source,
                                        globals_dict,
                                        free_vars,
                                        closure_cells, 
                                        filename = filename)
    if config.print_untyped_function:
      print "[ast_conversion] Translated %s into untyped function:\n%s" % (fn, repr(fundef))
                
  register_python_fn(original_fn, fundef)
  register_python_fn(fn, fundef)
  _currently_processing.remove(original_fn)
  return fundef


#####################################
#            DEBUG OUTPUT           #
#####################################

# show untyped IR after it's translated from Python?
print_untyped_function = False

# show the higher level typed function after specialization?
print_specialized_function = False

# print function after all adverbs have been turned to loops
print_loopy_function = False

# show lower level typed function before
# it gets translated to LLVM?
print_lowered_function = False

# show LLVM bytecode before optimization passes
print_unoptimized_llvm = False

# show LLVM bytecode after optimizations
print_optimized_llvm = False

# before starting function specialization, print the fn name and input types 
print_before_specialization = False

# show the input function to each transformation?
print_functions_before_transforms = [] #['LowerAdverbs'] # ['DCE'] 

# show the function produced by each transformation?
print_functions_after_transforms = [] # ['Fusion'] #['LowerAdverbs'] #['DCE'] 

# show aliases and escape sets
print_escape_analysis = False

# how long did each transform take?
print_transform_timings = False

# print each transform's name when it runs
print_transform_names = False

# at exit, print the names of all specialized functions
print_specialized_function_names = False

# print generated assembly of compiled functions
print_x86 = False

######################################
#        PARAKEET OPTIMIZATIONS      #
######################################

opt_verify = True
opt_inline = True
opt_fusion = True

opt_licm = True
opt_copy_elimination = True
opt_stack_allocation = True
opt_index_elimination = True
opt_range_propagation = True
opt_shape_elim = True
opt_scalar_replacement = True
opt_redundant_load_elimination = True  


# may dramatically increase compile time
opt_loop_unrolling = False

# recompile functions for distinct patterns of unit strides
# in array arguments
stride_specialization = True

######################################
#           LLVM OPTIONS             #
######################################

# run LLVM optimization passes
llvm_optimize = True

# number of times to run optimizations
llvm_num_passes = 4

# run verifier over generated LLVM code?
llvm_verify = True





from testing_helpers import expect, eq, run_local_tests
from parakeet import jit
import numpy as np

@jit
def min_(a,b):
  return min(a,b)

def test_min_int():
  expect(min_, [-3, 4], -3)

def test_min_float():
  expect(min_, [-3.0, 4.0], -3.0)

@jit
def max_(a,b):
  return max(a,b)

def test_max_int():
  expect(max_, [3,4], 4)

def test_max_float():
  expect(max_, [3.0,4.0], 4.0)

@jit
def sum_(xs):
  return sum(xs)

def test_sum_int():
  x = np.array([1,2,3])
  expect(sum_, [x], sum(x))

def test_sum_float():
  x = np.array([1.0,2.0,3.0])
  expect(sum_, [x], sum(x))

def sum_bool():
  x = np.array([True, False, True])
  expect(sum_, [x], sum(x))

@jit
def range1(stop):
  return range(stop)

def test_range1():
  expect(range1, [10], np.arange(10))

@jit
def range2(start, stop):
  return range(start, stop)

def test_range2():
  expect(range2, [10, 20], np.arange(10,20))

    
@jit
def range3(start, stop, step):
  return range(start, stop, step) 

def test_range3():
  expect(range3, [20,45,3], np.arange(20,45,3))

@jit 
def to_int(x):
  return int(x)

def test_to_int():
  expect(to_int, [1.2], 1)


@jit 
def to_float(x):
  return float(x)

def test_to_float():
  res = to_float(1)
  assert type(res) is np.float64, "Expected type float, got: %s" % (type(res),)
  assert res == 1.0

  res = to_float(True)
  assert res == 1.0

@jit
def to_long(x):
  return long(x)

def test_to_long():
  res = to_long(-1.0)
  assert res == -1.0

if __name__ == '__main__':
  run_local_tests()



import numpy as np
import parakeet
import testing_helpers

xs = [0.1, 0.5, 2.0]

def test_sin():
  testing_helpers.expect_each(parakeet.sin, np.sin, xs)

def test_sinh():
  testing_helpers.expect_each(parakeet.sinh, np.sinh, xs)

def test_cos():
  testing_helpers.expect_each(parakeet.cos, np.cos, xs)

def test_cosh():
  testing_helpers.expect_each(parakeet.cosh, np.cosh, xs)

def test_tan():
  testing_helpers.expect_each(parakeet.tan, np.tan, xs)

def test_tanh():
  testing_helpers.expect_each(parakeet.tanh, np.tanh, xs)

def test_log():
  testing_helpers.expect_each(parakeet.log, np.log, xs)

def test_log10():
  testing_helpers.expect_each(parakeet.log10, np.log10, xs)

def test_exp():
  testing_helpers.expect_each(parakeet.exp, np.exp, xs)

if __name__ == '__main__':
  testing_helpers.run_local_tests()



import time

class timer(object):
  def __init__(self, name = None, newline = True):
    self.name = name 
    self.start_t = time.time()
    self.newline = newline
    
  def __enter__(self):
    self.start_t = time.time()
  
  def elapsed(self):
    return time.time() - self.start_t
  
  def __exit__(self, exc_type, exc_value, traceback):
    t = self.elapsed()
    if self.newline:
      print 
    s = "Elapsed time: " if self.name is None else "%s : " % self.name 
    if exc_type is None:
      s += "%0.4f" % t
    else:
      name = str(exc_type) if exc_type.__name__ is None else exc_type.__name__
      s += "FAILED with %s '%s'" % (name, exc_value)
    print s 
    # don't raise exceptions
    return exc_type is not KeyboardInterrupt

from parakeet import jit
from numba import autojit
import numpy as np 

def compare_perf(fn, args, numba= True, cpython = True):
  
  parakeet_fn = jit(fn)
  name = fn.__name__
  parakeet_result = None
  numba_result = None 
  cpython_result = None 

  with timer('Parakeet #1 -- %s' % name):
    parakeet_result = parakeet_fn(*args)

  with timer('Parakeet #2 -- %s' % name):
    parakeet_result = parakeet_fn(*args)

  if numba:
    numba_fn = autojit(fn)

    with timer('Numba #1 -- %s' % name):
      numba_result = numba_fn(*args)

    with timer('Numba #2 -- %s' % name):
      numba_result = numba_fn(*args)
  
  if parakeet_result is not None and numba_result is not None:  
    assert np.allclose(parakeet_result, numba_result)  
  
  if cpython:
    with timer('Python -- %s' % name):
      python_result = fn(*args)

  if cpython_result is not None and parakeet_result is not None:
    assert np.allclose(parakeet_result, python_result)  



import llvm.core as llcore 
# from llvm.core import Type as lltype  

from core_types import ScalarT, FloatT, Int32, Int64
from llvm_types import llvm_value_type
 
def const(python_scalar, parakeet_type):
  assert isinstance(parakeet_type, ScalarT)
  llvm_type = llvm_value_type(parakeet_type)
  if isinstance(parakeet_type, FloatT):
    return llcore.Constant.real(llvm_type, float(python_scalar))
  else:
    return llcore.Constant.int(llvm_type, int(python_scalar))

def int32(x):
  """Make LLVM constants of type int32"""
  return const(x, Int32)

def int64(x):
  return const(x, Int64)

def zero(llvm_t):
  """
  Make a zero constant of either int or real type. 
  Doesn't (yet) work for vector constants! 
  """
  if isinstance(llvm_t, llcore.IntegerType):
    return llcore.Constant.int(llvm_t, 0)
  else:
    return llcore.Constant.real(llvm_t, 0.0)
  
def one(llvm_t):
  """
  Make a constant 1 of either int or real type. 
  Doesn't (yet) work for vector constants!
  """
  if isinstance(llvm_t, llcore.IntegerType):
    return llcore.Constant.int(llvm_t, 1)
  else:
    return llcore.Constant.real(llvm_t, 1.0)


import numpy as np


import config
import type_conv_decls
from decorators import jit, macro
from lib import *
from run_function import run, specialize_and_compile
import syntax 

def typed_repr(fn, args):
  _, typed, _, _ = specialize_and_compile(fn, args)
  return typed

def clear_specializations():
  import closure_type
  for clos_t in closure_type._closure_type_cache.itervalues():
    clos_t.specializations.clear()




from syntax import Assign, Index, Var
from loop_transform import LoopTransform 


class ScalarReplacement(LoopTransform):
  """
  When a loop reads and writes to a non-varying memory location, 
  we can keep the value of that location in a register and write 
  it to the heap after the loop completes. 
  
  Transform code like this:
      for i in low .. high: 
        z = x[const]
        q = z ** 2
        x[const] = q + i 
  into 
      z_in = x[const]
      for i in low .. high:
        (header)
          z_loop = phi(z_in, z_out)
        (body) 
          q = z_loop ** 2
          z_out = q + i 
      x[const] = z_loop   
  """
  
  
  
  def preload_reads(self, reads):
    scalar_vars = {}
    for (array_name, index_set) in reads.iteritems():
      t = self.type_env[array_name]
      for index_expr in index_set:
        scalar = self.index(Var(array_name, type = t), index_expr, 
                            temp = True, name = "scalar_repl_input")
       
        scalar_vars[(array_name, index_expr)] = scalar
    return scalar_vars

  def replace_indexing(self, loop_body, loop_scalars):
    """
    Given a map from (array_name, index_expr) pairs to 
    scalar variables, replace reads/writes with scalars
    """
    final_scalars = loop_scalars.copy()
    for stmt in loop_body:
      if stmt.__class__ is not Assign: 
        continue 
      if stmt.rhs.__class__ is Index:
        key = stmt.rhs.value.name, stmt.rhs.index
        if key in final_scalars:
          stmt.rhs = final_scalars[key]  
      if stmt.lhs.__class__ is Index:
        key = stmt.lhs.value.name, stmt.lhs.index
        if key in final_scalars:
          new_var = self.fresh_var(stmt.lhs.type, "scalar_repl_out")
          stmt.lhs = new_var
          final_scalars[key] = new_var  
    return final_scalars
  
  
  def transform_ForLoop(self, stmt):
    
    if self.is_simple_block(stmt.body):
      loop_vars = set([stmt.var.name])
      loop_vars.update(stmt.merge.keys())
      # gather all the variables whose values change between loop iters
      self.collect_loop_vars(loop_vars, stmt.body)
      # full set of array reads & writes 
      reads, writes = self.collect_memory_accesses(stmt.body)
      
      # which arrays are written to at loop-dependent locations? 
      unsafe = set([])
      for (name, write_indices) in writes.iteritems():
        if self.any_loop_vars(loop_vars, write_indices):
          unsafe.add(name)
      safe_writes = dict([(k,v) for (k,v) in writes.items() 
                          if k not in unsafe])
      
      safe_reads = dict([(k,v) for (k,v) in reads.items() 
                         if k not in unsafe and not self.any_loop_vars(loop_vars, v)])

      
      safe_locations = {}
      all_keys = set(safe_writes.keys()).union(set(safe_reads.keys()))
      for name in all_keys:
        index_set = safe_reads.get(name, set([])).union(safe_writes.get(name, set([])))
        safe_locations[name] = index_set 
      # move all safe/loop-invariant reads into registers at the top of the loop
      # I'm also including the writes (by passing safe_locations instead of safe_reads)
      # so that they become loop-carried accumulator values, otherwise where else would
      # they get their initial values?
      input_scalars = self.preload_reads(safe_locations)
      
      # need to rename the scalars so they have an SSA variable for the beginning of the loop
      loop_scalars = {}
      for ( (name,index_expr), input_var) in input_scalars.iteritems():
        loop_var = self.fresh_var(input_var.type, "scalar_repl_acc")
        loop_scalars[(name, index_expr)] = loop_var
      
      
      # propagate register names for all writes 
      final_scalars = self.replace_indexing(stmt.body, loop_scalars)
      
      for (key, final_var) in final_scalars.iteritems():
        loop_var = loop_scalars[key]
        input_var = input_scalars[key]
        stmt.merge[loop_var.name] = (input_var, final_var)
      
      self.blocks.append(stmt)
      # write out the results back to memeory 
      for ( (array_name, index_expr), loop_var) in loop_scalars.iteritems():
        array_type = self.type_env[array_name]
        lhs = self.index(Var(array_name, type = array_type), index_expr, temp = False)
        self.assign(lhs, loop_var) 
      return None
    else:     
      stmt.body = self.transform_block(stmt.body)
      return stmt 


from core_types import Int32, Int64, Float32, Float64
from loop_transform import LoopTransform

class Vectorize(LoopTransform):
  
  def pre_apply(self, _):
    # skip the may-alias analysis from LoopTransform
    pass 
  
  vector_elt_types = (Int32, Int64, Float32, Float64) 
  def vectorize_loop(self, stmt):
     
    candidate_names = set([])
    before_values = {}
    after_values = {}
    for (k, (before, after)) in stmt.merge.iteritems():
      t = before.type
      if t in self.vector_elt_types:
        candidate_names.add(k)
        before_values[k] = before
        after_values[k] = after
      
      
      
      
      
      
      
      
      
  
  def transform_ForLoop(self, stmt):
    # for now we only vectorize if 
    # something is being accumulated, 
    # otherwise you need more sophisticated
    # cost-base heuristic to avoid lots of 
    # redundant SIMD packing and unpacking
    if self.is_simple_block(stmt.body) and len(stmt.merge) > 1:
      return self.vectorize_loop(stmt)
    else:
      return stmt 
      


from array_type import ArrayT, SliceT
from closure_type import ClosureT
from core_types import NoneT, ScalarT, StructT, TypeValueT
from tuple_type import TupleT
from shape import Shape, Tuple, Closure, Var, Slice, Struct




def from_types(types):
  return Converter().from_types(types)

class Converter(object):
  """
  Turn a list of input types into a list of abstract values, 
  numbering the input arrays and scalars but preserving the 
  structure of tuples, closures, and slices
  """
  def __init__(self):
    self.counter = 0
  
  def fresh_var(self):
    n = self.counter 
    self.counter += 1
    return Var(n)
    
  def from_type(self, t):
    if isinstance(t, ScalarT):
      return self.fresh_var()
    elif t.__class__ is ArrayT:
      dim_vars = [self.fresh_var() for _ in range(t.rank)]
      return Shape(dim_vars)
    elif t.__class__ is TupleT:
      elt_values = self.from_types(t.elt_types)
      return Tuple(elt_values)
    elif t.__class__ is SliceT:
      start = self.from_type(t.start_type)
      stop = self.from_type(t.stop_type)
      step = self.from_type(t.step_type)
      return Slice(start, stop, step)
    elif t.__class__ is ClosureT:
      arg_vals = self.from_types(t.arg_types)
      return Closure(t.fn, arg_vals)
    elif isinstance(t, StructT):
      field_names = [fn for (fn,_) in t._fields_]
      field_types = [ft for (_,ft) in t._fields_]
      field_vals = self.from_types(field_types)
      return Struct(field_names, field_vals)
    elif isinstance(t, (TypeValueT, NoneT)):
      return Tuple(())
    else:
      assert False, "Unsupported type: %s" % t
    
  def from_types(self, arg_types):
    values = []
    for t in arg_types:
      v = self.from_type(t)
      values.append(v)
    return values



import syntax_helpers

from adverb_semantics import AdverbSemantics
from syntax import Index, Map 
from transform import Transform

class LowerAdverbs(AdverbSemantics, Transform):
  def transform_TypedFn(self, expr):
    import pipeline 
    return pipeline.loopify(expr)
    
  def transform_IndexMap(self, expr, output = None):
    fn = self.transform_expr(expr.fn)
    shape = self.transform_expr(expr.shape)
    return self.eval_index_map(fn, shape, output)
  
  def transform_IndexReduce(self, expr):
    fn = self.transform_expr(expr.fn)
    combine = self.transform_expr(expr.combine)
    shape = self.transform_expr(expr.shape)
    init = self.transform_if_expr(expr.init)
    return self.eval_index_reduce(fn, combine, shape, init)
    
  def transform_Map(self, expr, output = None):
    fn = self.transform_expr(expr.fn)
    args = self.transform_expr_list(expr.args)
    axis = syntax_helpers.unwrap_constant(expr.axis)
    return self.eval_map(fn, args, axis, output = output)

  def transform_Reduce(self, expr):
    fn = self.transform_expr(expr.fn)
    args = self.transform_expr_list(expr.args)
    combine = self.transform_expr(expr.combine)
    init = self.transform_if_expr(expr.init)
    axis = syntax_helpers.unwrap_constant(expr.axis)
    return self.eval_reduce(fn, combine, init, args, axis)

  def transform_Scan(self, expr):
    fn = self.transform_expr(expr.fn)
    args = self.transform_expr_list(expr.args)
    combine = self.transform_expr(expr.combine)
    print expr.emit 
    emit = self.transform_expr(expr.emit)

    init = self.transform_if_expr(expr.init)
    axis = syntax_helpers.unwrap_constant(expr.axis)
    return self.eval_scan(fn, combine, emit, init, args, axis)

  def transform_AllPairs(self, expr):
    fn = self.transform_expr(expr.fn)
    args = self.transform_expr_list(expr.args)
    assert len(args) == 2
    x,y = self.transform_expr_list(args)
    axis = syntax_helpers.unwrap_constant(expr.axis)
    return self.eval_allpairs(fn, x, y, axis)
  
  def transform_Assign(self, stmt):
    if stmt.lhs.__class__ is Index and isinstance(stmt.rhs, Map):
      self.transform_Map(stmt.rhs, output = stmt.lhs)
      return None 
    return Transform.transform_Assign(self, stmt)
  



import numpy as np
import testing_helpers
import parakeet 

int_vec = 300 + np.arange(300, dtype=int)
float_vec = int_vec.astype(float)
bool_vec = float_vec < np.mean(float_vec)

a = np.arange(500, dtype=float).reshape(50,10)
b = np.arange(100,600).reshape(50,10)

def my_sum(xs):
  return parakeet.reduce(parakeet.add, xs, init=0)

def test_int_sum():
  testing_helpers.expect(my_sum, [int_vec], np.sum(int_vec))

def test_float_sum():
  testing_helpers.expect(my_sum, [float_vec], np.sum(float_vec))

def test_bool_sum():
  testing_helpers.expect(my_sum, [bool_vec], np.sum(bool_vec))

def sqr_dist(y, x):
  return sum((x-y)*(x-y))

def reduce_2d(Ys):
  init = parakeet.zeros_like(Ys[0])
  return parakeet.reduce(parakeet.add, Ys, init = init, axis = 0)

def test_2d_reduce():
  par_rslt = reduce_2d(a)
  np_rslt = np.sum(a, 0)
  assert testing_helpers.eq(par_rslt, np_rslt), \
    "Expected %s but got %s" % (np_rslt, par_rslt)

def test_sqr_dist():
  z = a[0]
  def run_sqr_dist(c):
    return sqr_dist(z, c)
  par_rslt = parakeet.each(run_sqr_dist, a)
  py_rslt = np.array(map(run_sqr_dist, a))
  assert testing_helpers.eq(par_rslt, py_rslt), \
      "Expected %s but got %s" % (py_rslt, par_rslt)

def avg_along_axis_0(Xs):
  assign = np.array([0,0,1,0,1,0,1,0,1,1])
  Ys = Xs[assign == 1]

  def zero(x):
    return 0.0
  zeros = parakeet.each(zero, Xs[0])
  s = reduce(parakeet.add, Ys, init=zeros)
  def d(s):
    return s / Ys.shape[0]
  return parakeet.each(d, s)

if __name__ == '__main__':
  testing_helpers.run_local_tests()



import numpy as np 
import parakeet
import testing_helpers

def use_zip(x,y):
  return zip(x,y)

def test_zip_tuples():
  a = (1,2,3)
  b = (np.array([1.0]), np.array([2.0]), np.array([3.0]))
  testing_helpers.expect(use_zip, [a,b], zip(a,b))

if __name__ == '__main__':
  testing_helpers.run_local_tests() 




class NameNotFound(Exception):
  def __init__(self, name):
    self.name = name
    

  def __str__(self):
    return self.name 
  
versions = {}
original_names = {}
  
def get(name):
  version = versions.get(name)
  if version is None:
    raise NameNotFound(name)
  else:
    return "%s.%d" % (name, version)
    
def fresh(name):
  version = versions.get(name, 0) + 1 
  versions[name] = version
  if version == 1:
    ssa_name = name 
  else:
    ssa_name = "%s.%d" % (name, version)
  original_names[ssa_name] = name
  return ssa_name 

lcase_chars = [chr(i + 97) for i in xrange(26)]
def fresh_list(count):
  prefixes = lcase_chars[:count]
  while len(prefixes) < count:
    count -= 26
    prefixes += lcase_chars[:count]
  return map(fresh, prefixes)


def original(unique_name):
  original_name = original_names.get(unique_name)
  if original_name is None:
    return unique_name 
  else:
    return original_name 
  
def refresh(unique_name):
  """Given an existing unique name, create another versioned name with the same root"""
  try:
    return fresh(original(unique_name))
  except NameNotFound:
    # it wasn't really an SSA name but keep going anyway 
    return fresh(unique_name)


import prims 

from offset_analysis import OffsetAnalysis
from syntax import Var
from syntax_helpers import true, false 
from transform import Transform 

class RangePropagation(Transform):
  def pre_apply(self, fn):
    # map variables to pairs which include low/exclude high  
    self.known_ranges = {}
    self.known_offsets = OffsetAnalysis().visit_fn(fn)
  
  def transform_ForLoop(self, stmt):
    self.known_ranges[stmt.var.name] = (stmt.start, stmt.stop)
    stmt.body = self.transform_block(stmt.body)
    return stmt 
  
  def transform_PrimCall(self, expr):
    if expr.prim == prims.less_equal:
      x, y = expr.args
      
      if x.__class__ is Var:
        offsets = [(x.name, 0)]
        if x.name in self.known_offsets:
          offsets = self.known_offsets[x.name].union(set(offsets))
        for (var_name, offset) in offsets:
          if var_name in self.known_ranges:
            (low,high) = self.known_ranges[var_name]
            if y == high and offset >= 0:
              return false 
            elif y == low and offset <= 0:
              return true   
    return expr
      



class ValueMismatch(Exception):
  """
  Raise this exception whenever two incompatible abstract values get combined
  """

  def __init__(self, v1, v2):
    self.v1 = v1
    self.v2 = v2

  def __str__(self):
    return "ValueMismatch(%s, %s)" % (self.v1, self.v2)

  def __repr__(self):
    return str(self)

class AbstractValue(object):
  def combine(self, other):
    if self == other:
      return self
    else:
      return unknown_value
    
  def __repr__(self):
    return str(self)

class Unknown(object):
  """Bottom of the abstract shape lattice"""

  def __eq__(self, other):
    return other.__class__ is Unknown

  def combine(self, other):
    return other

  def __str__(self):
    return "<unknown>"

  def __repr__(self):
    return str(self)

unknown_value = Unknown()

class AnyValue(object):
  """Top of the abstract shape lattice"""

  def __eq__(self, other):
    return other.__class__ is AnyValue

  def combine(self, other):
    return self

  def __str__(self):
    return "<any>"

  def __repr__(self):
    return str(self)

any_value = AnyValue

class Scalar(AbstractValue):
  """Base class for all scalar operations"""

  def combine(self, other):
    if self == other:
      return self
    else:
      return any_scalar
  rank = 0

class AnyScalar(Scalar):
  def __eq__(self, other):
    return other.__class__ is AnyScalar

  def combine(self, other):
    assert isinstance(other, Scalar), \
        "Can't combine scalar with %s" % other
    return self

  def __str__(self):
    return "Scalar"

any_scalar = AnyScalar()

class Const(Scalar):
  def __init__(self, value):
    self.value = value

  def __eq__(self, other):
    return isinstance(other, Const) and other.value == self.value

  def __str__(self):
    return str(self.value)

  def combine(self, other):
    if self == other:
      return self
    elif isinstance(other, Scalar):
      return any_scalar
    else:
      raise ValueMismatch(self, other)

def const(x):
  return Const(x)

def is_zero(d):
  return d.__class__ is Const and d.value == 0

def is_one(d):
  return d.__class__ is Const and d.value == 1

def is_none(d):
  return d.__class__ is Const and d.value is None

class Binop(Scalar):
  def __init__(self, x, y):
    self.x = x
    self.y = y

  def __eq__(self, other):
    return self.__class__ == other.__class__ and \
           self.x == other.x and \
           self.y == other.y

  def __repr__(self):
    return str(self)

  def __hash__(self):
    return hash( (self.x, self.y) )

class Sub(Binop):
  def __str__(self):
    return "%s - %s" % (self.x, self.y)

class Add(Binop):
  def __str__(self):
    return "%s + %s" % (self.x, self.y)

class Div(Binop):
  def __str__(self):
    return "%s / %s" % (self.x, self.y)

class Mult(Binop):
  def __str__(self):
    return "%s * %s" % (self.x, self.y)

class Mod(Binop):
  def __str__(self):
    return "%s % %s" % (self.x, self.y)

class Var(Scalar):
  def __init__(self, num):
    self.num = num

  def __eq__(self, other):
    return other.__class__ is Var and\
           self.num == other.num

  def __hash__(self):
    return hash(self.num)

  def __str__(self):
    return "x%d" % self.num

  def __repr__(self):
    return str(self)

  def combine(self, other):
    if self == other:
      return self
    else:
      # combining two different variables returns an unknown scalar
      return any_scalar

class Shape(AbstractValue):
  def __init__(self, dims):
    assert len(dims) > 0
    self.dims = [const(d) if isinstance(d, int) else d for d in dims]
    self.rank = len(dims)

  def __eq__(self, other):
    return other.__class__ is  Shape and \
           len(self.dims) == len(other.dims) and \
           all(d1 == d2 for (d1,d2) in zip(self.dims, other.dims) )

  def __str__(self):
    return "Shape(%s)" % (", ".join(str(d) for d in self.dims))

  def __repr__(self):
    return str(self)

  def combine(self, other):
    if isinstance(other, Shape) and other.rank == self.rank:
      dims = combine_pairs(self.dims, other.dims)
      return make_shape(dims)
    raise ValueMismatch(self, other)

def make_shape(dims):
  if len(dims) == 0:
    return any_scalar
  return Shape(tuple(dims))

def dim(shape, d):
  if isinstance(shape, Shape):
    return shape.dims[d]
  else:
    # if the shape isn't available, getting the d'th
    # dimension returns an unknown scalar
    return any_scalar

def dim_list(shapes, d, exclude_scalars=False):
  if exclude_scalars:
    shapes = [s for s in shapes if not is_scalar(s)]
  return [dim(s,d) for s in shapes]

def array_of_unknown_shape(rank):
  return Shape([any_scalar] * rank)

def lower_rank(x, axis):
  assert isinstance(x, Shape), "Can't decrease rank of %s" % x
  # by convention, lowering a scalar returns a scalar
  if  axis >= x.rank or x.rank == 1:
    return any_scalar

  new_dims = []
  for (i,d) in enumerate(x.dims):
    if i != axis:
      new_dims.append(d)
  return make_shape(new_dims)

def lower_ranks(xs, axis):
  return [lower_rank(x, axis) for x in xs]

def increase_rank(x, axis, dim_expr):
  if isinstance(dim_expr, int):
    dim_expr = const(dim_expr)

  if isinstance(x, Shape):
    # we're taking dims d1...dm and constructing a
    # new shape d1...d_new...dm by inserting d_new
    # in the slot of the given axis
    assert axis <= x.rank
    if axis < len(x.dims):
      new_dims = []
      for (i,d) in enumerate(x.dims):
        if i == axis:
          new_dims.append(dim_expr)
        new_dims.append(d)
    else:
      new_dims = [d for d in x.dims]
      new_dims.append(dim_expr)
    return make_shape(new_dims)
  elif is_scalar(x):
    return Shape([dim_expr])
  else:
    raise RuntimeError("Don't know how to raise rank of value %s" % x)

def is_scalar(v):
  return isinstance(v, Scalar)

class ConstSlice(AbstractValue):
  def __init__(self, nelts):
    self.nelts = nelts

  def __str__(self):
    return "ConstSlice(nelts = %d)" % self.nelts

  def __eq__(self, other):
    return other.__class__ is ConstSlice and \
           other.nelts == self.nelts

  def combine(self, other):
    if other.__class__ is ConstSlice and other.nelts == self.nelts:
      return self
    else:
      return any_slice

class Slice(AbstractValue):
  def __init__(self, start, stop, step):
    self.start = start
    self.stop = stop
    self.step = step

  def __eq__(self, other):
    return other.__class__ is Slice and \
           self.start == other.start and \
           self.stop == other.stop and \
           self.step == other.step

  def __str__(self):
    return "Slice(%s, %s, %s)" % (self.start, self.stop, self.step)

  def combine(self, other):
    if isinstance(other, Slice):
      start = self.start.combine(other.start)
      stop = self.stop.combine(other.stop)
      step = self.step.combine(other.step)
      return Slice(start, stop, step)
    else:
      raise ValueMismatch(self, other)

any_slice = Slice(any_scalar, any_scalar, any_scalar)

class Tuple(AbstractValue):
  def __init__(self, elts):
    self.elts = tuple(elts)

  def __eq__(self, other):
    return other.__class__ is Tuple and \
           len(self.elts) == len(other.elts) and \
           all(e1 == e2 for (e1, e2) in zip(self.elts, other.elts))

  def __len__(self):
    return len(self.elts)

  def __iter__(self):
    return iter(self.elts)

  def __str__(self):
    return "Tuple(%s)" % ", ".join(str(e) for e in self.elts)

  def __getitem__(self, idx):
    return self.elts[idx]

  def combine(self, other):
    if other.__class__ is Tuple:
      if len(self.elts) == len(other.elts):
        return Tuple(combine_pairs(self.elts, other.elts))
    raise ValueMismatch(self, other)

class Closure(AbstractValue):
  def __init__(self, fn, args):
    self.fn = fn
    self.args = args

  def __str__(self):
    return "Closure(fn = %s, %s)" % \
           (self.fn, ", ".join(str(a) for a in self.args))

  def __eq__(self, other):
    return other.__class__ is Closure and \
           self.fn == other.fn and \
           len(self.arg_shapes) == len(other.arg_shapes) and \
           all(v1 == v2 for (v1,v2) in zip(self.args, other.args))

  def combine(self, other):
    if other.__class__ is  Closure:
      # TODO: Implement sets of closures like we have in the type system
      if self.fn == other.fn and \
         len(self.args) == len(other.args):
        combined_args = combine_pairs(self.args, other.args)
        return Closure(self.fn, combined_args)
    raise ValueMismatch(self, other)

class Struct(AbstractValue):
  def __init__(self, field_names, field_values):
    self.fields = field_names
    self.values = field_values

  def __str__(self):
    field_strings = ["%s = %s" % (k,v)
                     for (k,v) in zip(self.fields, self.values)]
    return "Struct(%s)" % ", ".join(field_strings)

  def __eq__(self, other):
      return other.__class__ is Struct and \
             len(self.fields) == len(other.fields) and \
             all(n1 == n2 for (n1,n2) in zip(self.fields, other.fields)) and \
             all(v1 == v2 for (v1, v2) in zip(self.values, other.values))

  def combine(self, other):
    if other.__class__ is  Struct and \
       len(self.fields) == len(other.fields) and \
       all(n1 == n2 for (n1,n2) in zip(self.fields, other.fields)):
      combined_args = combine_pairs(self.values, other.values)
      if any(old_val != new_val
             for (old_val, new_val) in zip(self.values, combined_args)):
        return Struct(self.fields, combined_args)
      else:
        return self
    raise ValueMismatch(self, other)

def combine_list(xs, preserve_const = True):
  acc = unknown_value
  for x in xs:
    acc = acc.combine(x)
  if not preserve_const and isinstance(acc, Const):
    acc = any_scalar
  return acc

def combine_pairs(xs, ys):
  return [xi.combine(yi) for (xi, yi) in zip(xs, ys)]

def computable_dim(d):
  c = d.__class__ 
  return c is Var or c is Const or \
    (isinstance(d, Binop) and computable_dim(d.x) and computable_dim(d.y))



import abc

class Ref(object):
  __meta__ = abc.ABCMeta

  @abc.abstractmethod
  def deref(self):
    pass

class GlobalValueRef(Ref):
  def __init__(self, value):
    self.value = value 
    
  def deref(self):
    return self.value 
  
  def __str__(self):
    return "GlobalValueRef(%s)" % (self.value,)
  
  def __repr__(self):
    return str(self)
  
    def __eq__(self, other):
      return isinstance(other, GlobalValueRef) and self.value is other.value 

class GlobalNameRef(Ref):
  def __init__(self, globals_dict, name):
    self.globals_dict = globals_dict
    self.name = name

  def __str__(self):
    return "GlobalNameRef(%s)" % self.name

  def __repr__(self):
    return str(self)

  def deref(self):
    return self.globals_dict[self.name]

  def __eq__(self, other):
    return isinstance(other, GlobalNameRef) and \
           self.globals_dict is other.globals_dict and \
           self.name == other.name

class ClosureCellRef(Ref):
  def __init__(self, cell, name):
    self.cell = cell
    self.name = name

  def deref(self):
    return self.cell.cell_contents

  def __str__(self):
    return "ClosureCellRef(%s)" % self.name

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    if isinstance(other, ClosureCellRef):
      try:
        return self.cell == other.cell
      except:
        return self.cell is other.cell
    return False



import numpy as np

from parakeet import scan, add
from testing_helpers import run_local_tests, expect, expect_each

int_1d = np.arange(5)
float_1d = np.arange(5, dtype='float')
int_2d = np.array([int_1d, int_1d, int_1d])
float_2d = np.array([float_1d, float_1d, float_1d])

def running_sum(x):
  return scan(add, x)

def test_scan_add_1d():
  expect_each(running_sum, np.cumsum, [int_1d, float_1d])

def loop_row_sums(x):
  n_rows = x.shape[0]
  y = np.zeros_like(x)
  y[0, :] = x[0, :]
  for i in xrange(1,n_rows):
    y[i, :] = y[i-1, :] + x[i, :]
  return y

def test_scan_add_2d():
  expect_each(running_sum, loop_row_sums, [int_2d, float_2d])

if __name__ == '__main__':
  run_local_tests()



import ctypes
import llvm.core as llcore

from llvm.core import Type as lltype

from core_types import ScalarT, PtrT, NoneT

void_t = lltype.void()
int1_t = lltype.int(1)
int8_t = lltype.int(8)
int16_t = lltype.int(16)
int32_t = lltype.int(32)
int64_t = lltype.int(64)

float32_t = lltype.float()
float64_t = lltype.double()

ptr_int8_t = lltype.pointer(int8_t)
ptr_int32_t = lltype.pointer(int32_t)
ptr_int64_t = lltype.pointer(int64_t)

def nbytes(t):
  if t.kind == llcore.TYPE_FLOAT:
    return 4
  elif t.kind == llcore.TYPE_DOUBLE:
    return 8
  else:
    return t.width / 8

def is_scalar(llvm_t):
  return isinstance(llvm_t, llcore.IntegerType) or \
         llvm_t in (float32_t, float64_t)

_ctypes_scalars_to_llvm_types = {
  ctypes.c_bool : int8_t,
  ctypes.c_uint8 : int8_t,
  ctypes.c_int8 : int8_t,

  ctypes.c_uint16 : int16_t,
  ctypes.c_int16 : int16_t,

  ctypes.c_uint32 : int32_t,
  ctypes.c_int32 : int32_t,

  ctypes.c_uint64 : int64_t,
  ctypes.c_int64 : int64_t,

  ctypes.c_float : float32_t,
  ctypes.c_double : float64_t,
}

PyCPointerType = type(ctypes.POINTER(ctypes.c_int32))
PyCStructType = type(ctypes.Structure)

_struct_cache = {}
def ctypes_struct_to_lltype(S, name = None):
  if not name:
    name = S.__class__.__name__

  key = tuple(S._fields_)

  if key in _struct_cache:
    return _struct_cache[key]
  else:
    llvm_field_types = [ctypes_to_lltype(field_type)
                        for (_, field_type) in S._fields_]
    llvm_struct = lltype.struct(llvm_field_types, name)
    _struct_cache[key] = llvm_struct
    return llvm_struct

def ctypes_scalar_to_lltype(ct):
  assert ct in _ctypes_scalars_to_llvm_types, \
      "%s isn't a convertible to an LLVM scalar type" % ct
  return _ctypes_scalars_to_llvm_types[ct]

def ctypes_to_lltype(ctypes_repr, name = None):
  if type(ctypes_repr) == PyCStructType:
    return ctypes_struct_to_lltype(ctypes_repr, name)
  elif type(ctypes_repr) == PyCPointerType:
    elt_t = ctypes_repr._type_
    if elt_t == ctypes.c_bool:
      return lltype.pointer(int8_t)
    else:
      return lltype.pointer(ctypes_to_lltype(elt_t))
  else:
    return ctypes_scalar_to_lltype(ctypes_repr)

def llvm_value_type(t):
  return ctypes_to_lltype(t.ctypes_repr, t.node_type())

def llvm_ref_type(t):
  llvm_value_t = llvm_value_type(t)
  if isinstance(t, (PtrT, ScalarT, NoneT)):
    return llvm_value_t
  else:
    return lltype.pointer(llvm_value_t)

# we allocate heap slots for output scalars before entering the
# function
def to_llvm_output_type(t):
  llvm_type = llvm_value_type(t)
  if isinstance(t, ScalarT):
    return lltype.pointer(llvm_type)
  else:
    return llvm_type



import numpy as np
import testing_helpers
import parakeet 

from parakeet.interp import adverb_evaluator as interp

vec = np.array([1,4,9,16])
mat = np.array([vec, vec+100, vec+200, vec + 300])

expected_sqrt_vec = np.sqrt(vec)
expected_sqrt_mat = np.sqrt(mat)

def test_map_1d():
  sqrt_vec = interp.eval_map(np.sqrt, values=[vec], axis=0)
  assert testing_helpers.eq(sqrt_vec, expected_sqrt_vec), \
      "Expected %s from Map but got %s" % (expected_sqrt_vec, sqrt_vec)

def test_map_2d():
  sqrt_mat = interp.eval_map(np.sqrt, values=[mat], axis=0)
  assert testing_helpers.eq(sqrt_mat, expected_sqrt_mat), \
      "Expected %s from Map but got %s" % (expected_sqrt_mat, sqrt_mat)

"""
def sqrt_with_output(x, output):
  output[:] = np.sqrt(x)

def test_map_2d_output():
  out = np.zeros_like(expected_sqrt_mat)
  interp.eval_map(sqrt_with_output, values=[mat], axis=0, output=out)
  assert testing_helpers.eq(out, expected_sqrt_mat), \
      "Expected %s from Map with output param but got %s" % \
      (expected_sqrt_mat, out)
"""

def two_first_elts(x):
    return [x[0], x[0]]

"""
def two_first_elts_with_output(x, out):
    out[:] = two_first_elts(x)

def test_complex_map_2d_output():
  expected = np.array([two_first_elts(x) for x in mat])
  out = np.zeros_like(expected)
  interp.eval_map(two_first_elts_with_output, values=[mat], axis=0, output=out)
  assert testing_helpers.eq(out, expected), \
      "Expected %s from Map with output param but got %s" % \
      (expected, out)
"""

expected_sum_vec = np.sum(vec)

def test_reduce_1d():
  vec_sum = interp.eval_reduce(map_fn = interp.identity_function,
                               combine = np.add,
                               init = 0,
                               values = [vec],
                               axis = 0)

  assert testing_helpers.eq(vec_sum, expected_sum_vec), \
      "Expected %s from Reduce but got %s" % (expected_sum_vec, vec_sum)

expected_sum_mat = np.sum(mat, axis=0)

def test_reduce_2d():
  mat_sum = interp.eval_reduce(map_fn = interp.identity_function,
                               combine = np.add,
                               init = 0,
                               values = [mat],
                               axis = 0)

  assert testing_helpers.eq(mat_sum, expected_sum_mat), \
      "Expected %s from Reduce but got %s" % (expected_sum_mat, mat_sum)

"""
def add_vec_with_output(x, y, out):
    out[:] = x + y

def test_reduce_2d_output():
  output = np.zeros_like(expected_sum_mat)
  interp.eval_reduce(
    map_fn = interp.identity_function,
    combine = add_vec_with_output,
    init = 0,
    values = [mat],
    axis = 0,
    output = output)

  assert testing_helpers.eq(output, expected_sum_mat), \
      "Expected %s from Reduce (with output) but got %s" % \
      (expected_sum_mat, output)
"""
bool_vec = np.array([True, False, True, False, True])

def test_bool_sum():
  vec_sum = interp.eval_reduce(
    map_fn = interp.identity_function,
    combine = (lambda x,y: x + y),
    init = 0,
    values = [bool_vec],
    axis = 0)
  assert vec_sum == np.sum(bool_vec), \
      "Expected %s but got %s" % (np.sum(bool_vec), vec_sum)

expected_cumsum_vec = np.cumsum(vec)

def test_scan_1d():
  vec_prefixes = interp.eval_scan(map_fn=interp.identity_function,
                                  combine=np.add,
                                  emit=interp.identity_function,
                                  init=0,
                                  values=[vec],
                                  axis=0)

  assert testing_helpers.eq(vec_prefixes, expected_cumsum_vec), \
      "Expected %s from Scan but got %s" % (expected_cumsum_vec, vec_prefixes)

expected_cumsum_mat = np.cumsum(mat, axis=0)

def test_scan_2d():
  mat_prefixes = interp.eval_scan(
    map_fn = interp.identity_function,
    combine = np.add,
    emit = interp.identity_function,
    init = 0,
    values = [mat],
    axis = 0)

  assert testing_helpers.eq(mat_prefixes, expected_cumsum_mat), \
      "Expected %s from Scan but got %s" % (expected_cumsum_mat, mat_prefixes)

def test_allpairs():
  times_table = interp.eval_allpairs(np.multiply, vec, vec, 0)
  np_times_table = np.multiply.outer(vec, vec)
  assert testing_helpers.eq(times_table, np_times_table), \
      "Expected %s for AllPairs but got %s" % \
      (np_times_table, times_table)

  inner_products = interp.eval_allpairs(np.dot, mat, mat.T, 0)
  np_inner_products = np.dot(mat, mat)
  assert testing_helpers.eq(inner_products, np_inner_products), \
      "Expected %s for AllPairs but got %s" % \
      (inner_products, np_inner_products)

def test_index_map():
  indices = interp.eval_index_map(interp.identity_function, (10,))
  assert testing_helpers.eq(indices, np.arange(10)), \
    "Expected %s from IndexMap but got %s" % (np.arange(10), indices)
    
def test_index_reduce():
  n = 10 
  total = interp.eval_index_reduce(interp.identity_function, np.add, (n,), 0)
  expected = sum(np.arange(10,))
  assert total == expected, "Expected %d but got %d" % (expected, total)
  total_no_tuple = interp.eval_index_reduce(interp.identity_function, np.add, n, 0)
  assert total_no_tuple == expected, "Expected %d but got %d" % (expected, total_no_tuple )

if __name__ == '__main__':
  testing_helpers.run_local_tests()



import names
import syntax
import syntax_visitor

from subst import subst_stmt_list
from syntax import If, Assign, While, Return, ForLoop, Comment
from syntax import Var, TypedFn, Const, ExprStmt
from transform import Transform

class FoundCall(Exception):
  pass

class ContainsCalls(syntax_visitor.SyntaxVisitor):
  def visit_Call(self, expr):
    raise FoundCall

  def visit_fn(self, fn):
    try:
      self.visit_block(fn.body)
    except FoundCall:
      return True
    return False

def contains_calls(fn):
  return ContainsCalls().visit_fn(fn)

def replace_returns(stmts, output_var):
  """Change any returns at the outer scope into assignments to the output var"""

  new_stmts = []
  for stmt in stmts:
    c = stmt.__class__
    if c is Return:
      if output_var:
        new_stmts.append(syntax.Assign(output_var, stmt.value))
      continue
    if c is If:
      stmt.true = replace_returns(stmt.true, output_var)
      stmt.false = replace_returns(stmt.false, output_var)
    elif c in (ForLoop, While):
      stmt.body = replace_returns(stmt.body, output_var)

    new_stmts.append(stmt)
  return new_stmts

def can_inline_block(stmts, outer = False):
  for stmt in stmts:
    stmt_class = stmt.__class__
    if stmt_class is Assign or stmt_class is ExprStmt:
      pass
    elif stmt_class is If:
      return can_inline_block(stmt.true) and can_inline_block(stmt.false)
    elif stmt_class in (While, ForLoop):
      if not can_inline_block(stmt.body):
        return False
    elif stmt_class is Comment:
      continue
    else:
      assert stmt_class is Return, "Unexpected statement: %s" % stmt
      if not outer:
        return False
  return True

def can_inline(fundef):
  return can_inline_block(fundef.body, outer = True)

def replace_return_with_var(body, type_env, return_type):
  result_name = names.fresh("result")
  type_env[result_name] = return_type
  result_var = Var(result_name, type = return_type)
  new_body = replace_returns(body, result_var)
  return result_var, new_body

def do_inline(src_fundef, args, dest_type_env, dest_block, result_var = None):
  arg_names = src_fundef.arg_names
  n_expected = len(arg_names)
  n_given = len(args)
  arg_str = ",".join(src_fundef.arg_names)
  assert n_expected == n_given, \
      "Function %s expects %d args (%s) but given %d (%s)" % \
      (src_fundef.name, n_expected, arg_str, n_given, args)

  rename_dict = {}

  def rename_var(old_name):
    t = src_fundef.type_env[old_name]
    new_name = names.refresh(old_name)
    new_var = Var(new_name, type = t)
    rename_dict[old_name] = new_var
    dest_type_env[new_name] = t

    return new_var

  for (old_formal, actual) in zip(arg_names, args):
    if actual.__class__ in (Var, Const):
      rename_dict[old_formal] = actual
    else:
      new_var = rename_var(old_formal)
      dest_block.append(Assign(new_var, actual))

  for old_name in src_fundef.type_env.iterkeys():
    if old_name not in arg_names:
      rename_var(old_name)

  new_body = subst_stmt_list(src_fundef.body, rename_dict)
  if result_var is None:
    result_var, new_body = \
        replace_return_with_var(new_body, dest_type_env, src_fundef.return_type)
  else:
    new_body = replace_returns(new_body, result_var)
  dest_block.extend(new_body)
  return result_var

class Inliner(Transform):
  def __init__(self):
    Transform.__init__(self)
    self.count = 0

  def transform_TypedFn(self, expr):
    if self.fn.copied_by is not None:
      return self.fn.copied_by.apply(expr)
    else:
      # at the very least, apply high level optimizations
      import pipeline
      return pipeline.high_level_optimizations.apply(expr)

  def transform_Call(self, expr):
    target = self.transform_expr(expr.fn)
    closure_args = self.closure_elts(target)
    target = self.get_fn(target)
    if target.__class__ is TypedFn:
      if not can_inline(target):
        print "[Warning] Can't inline %s" % target
        return expr  
      self.count += 1
      curr_block = self.blocks.current()
      combined_args = tuple(closure_args) + tuple(expr.args)
      result_var = do_inline(target, combined_args,
                             self.type_env, curr_block)
      return result_var
    else:
      return expr

  def apply(self, fn):
    if contains_calls(fn):
      return Transform.apply(self, fn)
    else:
      return fn





import prims
import llvm.core as llc   

  

signed_int_comparisons = {
  prims.equal : llc.ICMP_EQ, 
  prims.not_equal : llc.ICMP_NE, 
  prims.greater : llc.ICMP_SGT, 
  prims.greater_equal : llc.ICMP_SGE, 
  prims.less : llc.ICMP_SLT, 
  prims.less_equal : llc.ICMP_SLE
}

unsigned_int_comparisons = {
  prims.equal : llc.ICMP_EQ, 
  prims.not_equal : llc.ICMP_NE, 
  prims.greater : llc.ICMP_UGT, 
  prims.greater_equal : llc.ICMP_UGE, 
  prims.less : llc.ICMP_ULT, 
  prims.less_equal : llc.ICMP_ULE
}  

float_comparisons = { 
  prims.equal : llc.FCMP_OEQ, 
  prims.not_equal : llc.FCMP_ONE, 
  prims.greater : llc.FCMP_OGT, 
  prims.greater_equal : llc.FCMP_OGE, 
  prims.less : llc.FCMP_OLT, 
  prims.less_equal : llc.FCMP_OLE                     
}


signed_binops = { 
  prims.add : 'add',  
  prims.subtract : 'sub', 
  prims.multiply : 'mul', 
  prims.divide : 'sdiv',
}

unsigned_binops = { 
  prims.add : 'add',  
  prims.subtract : 'sub', 
  prims.multiply : 'mul', 
  prims.divide : 'udiv', 
  prims.mod : 'urem', 
}

float_binops = { 
  prims.add : 'fadd',  
  prims.subtract : 'fsub', 
  prims.multiply : 'fmul', 
  prims.divide : 'fdiv', 
}

# Note: there is no division instruction between booleans
# so b1 / b2 should be translated to int(b1) / int(b2) 
bool_binops = { 
  prims.add : 'or_',
  prims.multiply : 'and_', 
  prims.subtract : 'xor',
  prims.divide : 'and_',
  
  prims.mod : 'urem'
  #prims.logical_not : 'not_'
}


import core_types 
import llvm_types

from llvm_context import global_context

float32_fn_t = llc.Type.function(llvm_types.float32_t, [llvm_types.float32_t])
float64_fn_t = llc.Type.function(llvm_types.float64_t, [llvm_types.float64_t])

def float32_fn(name):
  return global_context.module.add_function(float32_fn_t, name)

def float64_fn(name):
  return global_context.module.add_function(float64_fn_t, name)

float_unary_ops_list = [ 
  prims.tan, prims.tanh, prims.cos, prims.cosh, prims.sin, prims.sinh, 
  prims.log, prims.log10, prims.sqrt,
  prims.exp 
]

float_unary_ops = {}

def get_float_unary_op(prim, t):
  key = (prim, t)
  if key in float_unary_ops:
    return float_unary_ops[key]
  assert prim in float_unary_ops_list, \
    "Unsupported float primitive %s" % prim 
  assert t in (core_types.Float32, core_types.Float64), \
    "Invalid type %s, expected Float32 or Float64" % t
  fn_t = float32_fn_t if t == core_types.Float32 else float64_fn_t    
  llvm_value = global_context.module.add_function(fn_t, prim.name)
  float_unary_ops[key] = llvm_value 
  return llvm_value 



from syntax_expr import Expr 


#############################################################################
#
#  Array Operators: It's not scalable to keep adding first-order operators
#  at the syntactic level, so eventually we'll need some more extensible
#  way to describe the type/shape/compilation semantics of array operators
#
#############################################################################

class Len(Expr):
  _members = ['value']

class ConstArray(Expr):
  _members = ['shape', 'value']

class ConstArrayLike(Expr):
  """
  Create an array with the same shape as the first arg, but with all values set
  to the second arg
  """

  _members = ['array', 'value']

class Range(Expr):
  _members = ['start', 'stop', 'step']

class AllocArray(Expr):
  """Allocate an unfilled array of the given shape and type"""
  _members = ['shape', 'elt_type']
  
  def children(self):
    yield self.shape


class ArrayView(Expr):
  """Create a new view on already allocated underlying data"""

  _members = ['data', 'shape', 'strides', 'offset', 'size']

  def children(self):
    yield self.data
    yield self.shape
    yield self.strides
    yield self.offset
    yield self.size

class Ravel(Expr):
  _members = ['array']
  
  def children(self):
    return (self.array,)

class Reshape(Expr):
  _members = ['array', 'shape']
  
  def children(self):
    yield self.array 
    yield self.shape

class Shape(Expr):
  _members = ['array']
  
class Strides(Expr):
  _members = ['array']
    
class Transpose(Expr):
  _members = ['array']
  
  def children(self):
    yield self.array 
    
class Where(Expr):
  """
  Given a boolean array, returns its true indices 
  """
  _members = ['array']
  
  def children(self):
    yield self.array 



import numpy as np
import time 
from parakeet import jit 
from testing_helpers import run_local_tests, expect, eq

int_mat = np.reshape(np.arange(9), (3,3))
float_mat = np.sqrt(int_mat)
bool_mat = int_mat % 2

matrices = [int_mat, float_mat, bool_mat]

@jit 
def mm(X,Y,Z):
  m,d = X.shape
  n = Y.shape[1]
    
  for i in range(m):
      for j in range(n):
          total = 0
          for k in range(d):
              total += X[i,k] * Y[k,j]
          Z[i,j] = total 
  return Z

def test_loop_matmult():
  for X in matrices:
    for Y in matrices:
      res = np.dot(X, Y)
      Z = np.zeros(res.shape, dtype = res.dtype)
      expect(mm, [X,Y,Z], res)

if __name__ == '__main__':
    run_local_tests()



import core_types 
from core_types import Int64 as i64
  
class cfn:
  def __init__(self, name, arg_types, return_type):
    self.name = name 
    self.arg_types = arg_types 
    self.return_type = return_type 
    
pyobj = None 
c_str = None 

class PyDict:
  GetItem = cfn("PyDict_GetItem", [pyobj, pyobj], pyobj)
  GetItemString = cfn("PyDict_GetItemString", [pyobj, c_str], pyobj)
  SetItem = cfn("PyDict_SetItem", [pyobj, pyobj, pyobj], i64)
  SetItemString = cfn("PyDict_SetItemString", [pyobj, c_str,pyobj], i64)

class PyLong:
  pass 



"""  
("PyLong_AsDouble", f64type, [pyobj], []),
("PyLong_AsLong", i64type, [pyobj], []),
("PyLong_AsLongAndOverflow", i64type, [pyobj,i64type], []),
("PyLong_AsLongLong", i64type, [pyobj], []),
("PyLong_AsLongLongAndOverflow", i64type, [pyobj,i64type], []),
("PyLong_AsSsize_t", i64type, [pyobj], []),
("PyLong_AsUnsignedLong", i64type, [pyobj], []),
("PyLong_AsUnsignedLongLong", i64type, [pyobj], []),
("PyLong_AsUnsignedLongLongMask", i64type, [pyobj], []),
("PyLong_AsUnsignedLongMask", i64type, [pyobj], []),
("PyLong_AsVoidPtr", i8ptrtype, [pyobj], []),
("PyLong_FromDouble", pyobj, [f64type], []),
("PyLong_FromLong", pyobj, [i64type], []),
("PyLong_FromLongLong", pyobj, [i64type], []),
("PyLong_FromSize_t", pyobj, [i64type], []),
("PyLong_FromSsize_t", pyobj, [i64type], []),
("PyLong_FromString", pyobj, [i8ptrtype,i8ptrtype,i64type], []),
("PyLong_FromUnsignedLong", pyobj, [i64type], []),
("PyLong_FromUnsignedLongLong", pyobj, [i64type], []),
("PyLong_FromVoidPtr", pyobj, [i8ptrtype], []),
"""